import math

from casadi import cos, sin, tan
from do_mpc.controller import MPC
from do_mpc.model import Model
from do_mpc.simulator import Simulator
from scipy.spatial.distance import euclidean

from rrt_ml.deps.libraries.python_robotics_reeds_shepp import *
from rrt_ml.utilities.configs import *
from rrt_ml.utilities.formulas import *
from rrt_ml.utilities.hints import *
from rrt_ml.utilities.maps import *


# ORIGINAL (DO NOT CHANGE)
class MPCRS:
    """
    Model predictive controller with Redd-Shepp paths.
    """

    def __init__(self, cfg: MasterConfig):
        """
        Initialize.
        """

        self.cfg = cfg

        self.controller = None  # type: MPC | None
        self.desired_goal = None  # type: np.ndarray | None
        self.rs_xs = None  # type: list[float, ...] | None
        self.rs_ys = None  # type: list[float, ...] | None
        self.rs_yaws = None  # type: list[float, ...] | None
        self.rs_idx = None  # type: int | None
        self.tvp_template = None

        self._setup()

    def _setup(self):
        """
        Set up MPC.
        """

        # Set arbitrary value so that 'get_action' calls '_set_new_target' first time
        self.desired_goal = np.array([1000.0, 1000.0, 1000.0, 1000.0])

    def get_action(self, obs: dict[str, Vector4 | Vector6]):
        """
        Get env action with MPC controller.
        :param obs: current env observation for goal conditioned tasks
        :return: action vector
        """

        # If desired goal changed, we need to set a new target for the MPC
        if np.all(self.desired_goal != obs['desired_goal']):
            self._set_new_target(obs)
            self.controller.x0 = state4_to_pose3(obs['achieved_goal']).reshape(-1, 1)
            self.simulator.x0 = state4_to_pose3(obs['achieved_goal']).reshape(-1, 1)
            self.controller.set_initial_guess()

        # Get current car and target pose
        car_pose = state6_to_pose3(obs['observation'])
        target_pose = [self.rs_xs[self.rs_idx], self.rs_ys[self.rs_idx], self.rs_yaws[self.rs_idx]]

        # Check distance between car and target positions
        distance = euclidean(car_pose[0:2], target_pose[0:2])

        # If distance is small increase index, changing next target pose (tvp function depends on index) for the MPC
        if distance < self.cfg.rrt.rs.tolerance:

            # Check to go over the path length
            if self.rs_idx < (len(self.rs_xs) - 1):
                self.rs_idx += 1

        # Make state column vector
        state = car_pose.reshape(-1, 1)

        # Get control
        u = self.controller.make_step(state).flatten()
        self.simulator.make_step(u.reshape(-1, 1))

        # Now we need to convert to [-1, 1] range
        v_ref = scale_to_range(u[0], [-self.cfg.env.car.v_max, +self.cfg.env.car.v_max], [-1, 1])
        phi_ref = scale_to_range(u[1], [-self.cfg.env.car.phi_max, +self.cfg.env.car.phi_max], [-1, 1])

        return np.array([v_ref, phi_ref])

    def get_distance(self, node_from: Node, node_to: Node):
        """
        Calculate distance between nodes using RS path's length.
        :param node_from: start node
        :param node_to: final node
        :return: distance
        """

        # Get poses
        pose_init = node_from.pose
        pose_final = node_to.pose

        # Get RS path
        curvature = self.cfg.rrt.rs.curvature
        step_size = self.cfg.rrt.rs.step_size
        xs, ys, yaws, modes, lengths = reeds_shepp_path_planning(*pose_init, *pose_final, curvature, step_size)

        # Handle cases where could not calculate route
        if xs is None:
            lengths = [np.inf]

        return np.sum(np.abs(lengths))

    def _set_new_target(self, obs: dict):
        """
        Function to be called when setting a new desired goal: sets RS paths and function to update MPC cost.
        :param obs: gym env observation
        """

        # Update desired goal so that this set up is not called again
        self.desired_goal = obs['desired_goal']

        # Get current car and target pose to calculate waypoints
        car_pose = state6_to_pose3(obs['observation'])
        target_pose = state4_to_pose3(obs['desired_goal'])

        # Get list of waypoints
        curvature = self.cfg.rrt.rs.curvature
        step_size = self.cfg.rrt.rs.step_size
        xs, ys, yaws, modes, lengths = reeds_shepp_path_planning(*car_pose, *target_pose, curvature, step_size)
        self.rs_xs = xs[1:]
        self.rs_ys = ys[1:]
        self.rs_yaws = yaws[1:]

        # Initialize index
        self.rs_idx = 0

        # Get constants from config
        axis_dist = self.cfg.env.car.axis_dist
        n_horizon = self.cfg.rrt.mpc.n_horizon
        n_robust = self.cfg.rrt.mpc.n_robust
        step_size = self.cfg.rrt.mpc.step_size
        v_max = self.cfg.env.car.v_max
        v_min = -v_max
        phi_max = self.cfg.env.car.phi_max
        phi_min = -phi_max
        control_cost = self.cfg.rrt.mpc.control_cost

        # Initialize dynamics model
        model_type = 'continuous'  # either 'discrete' or 'continuous'
        model = Model(model_type)

        # Set state variables
        model.set_variable('_x', 'x')
        model.set_variable('_x', 'y')
        theta = model.set_variable('_x', 'theta')

        # Set control variables
        v = model.set_variable(var_type='_u', var_name='v')
        phi = model.set_variable(var_type='_u', var_name='phi')

        # Time varying parameters
        model.set_variable(var_type='_tvp', var_name='rs_x')
        model.set_variable(var_type='_tvp', var_name='rs_y')
        model.set_variable(var_type='_tvp', var_name='rs_yaw')

        # Set equations
        model.set_rhs('x', v * cos(theta))
        model.set_rhs('y', v * sin(theta))
        model.set_rhs('theta', v * tan(phi) / axis_dist)

        # Setup
        model.setup()

        # Initialize optimizer
        mpc = MPC(model)
        mpc.set_param(n_horizon=n_horizon, t_step=step_size, n_robust=n_robust, store_full_solution=True,
                      nlpsol_opts={'ipopt.print_level': 0, 'ipopt.sb': 'yes','print_time': 0})

        # Set state variables bound
        mpc.bounds['lower', '_x', 'theta'] = -2 * np.pi
        mpc.bounds['upper', '_x', 'theta'] = +2 * np.pi

        # Set control variables bound
        mpc.bounds['lower', '_u', 'v'] = v_min
        mpc.bounds['upper', '_u', 'v'] = v_max
        mpc.bounds['lower', '_u', 'phi'] = phi_min
        mpc.bounds['upper', '_u', 'phi'] = phi_max

        # Set cost for control input (avoid oscillations)
        mpc.set_rterm(v=control_cost, phi=control_cost)

        # Get template and define function
        self.tvp_template = mpc.get_tvp_template()

        def tvp_fun(t_now):
            if len(self.rs_xs[self.rs_idx:]) <= n_horizon:
                for k in range(n_horizon + 1):
                    self.tvp_template['_tvp', k, 'rs_x'] = self.rs_xs[-1]
                    self.tvp_template['_tvp', k, 'rs_y'] = self.rs_ys[-1]
                    self.tvp_template['_tvp', k, 'rs_yaw'] = self.rs_yaws[-1]
            else:
                for k in range(n_horizon + 1):
                    self.tvp_template['_tvp', k, 'rs_x'] = self.rs_xs[self.rs_idx]
                    self.tvp_template['_tvp', k, 'rs_y'] = self.rs_ys[self.rs_idx]
                    self.tvp_template['_tvp', k, 'rs_yaw'] = self.rs_yaws[self.rs_idx]
            return self.tvp_template

        # Set tvp fun
        mpc.set_tvp_fun(tvp_fun)

        # Finally, we need to add the m and l terms to cost function, but we do it when settings the target pose
        self.controller = mpc

        # Retrieve symbolic variables from model
        x = self.controller.model._x['x']
        y = self.controller.model._x['y']
        theta = self.controller.model._x['theta']
        x_target = self.controller.model.tvp['rs_x']
        y_target = self.controller.model.tvp['rs_y']
        theta_target = self.controller.model.tvp['rs_yaw']

        # Set cost for 'm' and 'l' term
        wp = self.cfg.rrt.mpc.weight_pos
        wa = self.cfg.rrt.mpc.weight_angle
        mterm = wp * (x - x_target) ** 2 + wp * (y - y_target) ** 2 + wa * (theta - theta_target) ** 2
        lterm = wp * (x - x_target) ** 2 + wp * (y - y_target) ** 2 + wa * (theta - theta_target) ** 2
        self.controller.set_objective(mterm=mterm, lterm=lterm)

        # Call setup to finalize
        self.controller.setup()

        # Set simulator
        self.simulator = Simulator(model)
        self.simulator.set_param(t_step=step_size)

        # TVP
        tvp_sim = self.simulator.get_tvp_template()

        def tvp_sim_fun(t_now):
            tvp_sim['rs_x'] = self.rs_xs[self.rs_idx]
            tvp_sim['rs_y'] = self.rs_ys[self.rs_idx]
            tvp_sim['rs_yaw'] = self.rs_yaws[self.rs_idx]
            return tvp_sim

        self.simulator.set_tvp_fun(tvp_sim_fun)
        self.simulator.setup()

    def _set_rs_path(self, obs: dict):
        """
        Set new RS path.
        :param obs: env observation
        """

        # Update desired goal so that this set up is not called again
        self.desired_goal = obs['desired_goal']

        # Get current car and target pose to calculate waypoints
        car_pose = state6_to_pose3(obs['observation'])
        target_pose = state4_to_pose3(obs['desired_goal'])

        # Get list of waypoints
        curvature = self.cfg.rrt.rs.curvature
        step_size = self.cfg.rrt.rs.step_size
        xs, ys, yaws, modes, lengths = reeds_shepp_path_planning(*car_pose, *target_pose, curvature, step_size)
        self.rs_xs = xs[1:]
        self.rs_ys = ys[1:]
        self.rs_yaws = yaws[1:]

        # Initialize index
        self.rs_idx = 0

# TEST
# class MPCRS:
#     """
#     Model predictive controller with Redd-Shepp paths.
#     """
#
#     def __init__(self, cfg: MasterConfig):
#         """
#         Initialize.
#         """
#
#         self.cfg = cfg
#
#         self.controller = None  # type: MPC | None
#         self.desired_goal = None  # type: np.ndarray | None
#         self.rs_xs = None  # type: list[float, ...] | None
#         self.rs_ys = None  # type: list[float, ...] | None
#         self.rs_yaws = None  # type: list[float, ...] | None
#         self.rs_idx = None  # type: int | None
#         self.tvp_template = None
#
#         self._setup()
#
#     def _setup(self):
#         """
#         Set up MPC.
#         """
#
#         # Set arbitrary value so that 'get_action' calls '_set_new_target' first time
#         self.desired_goal = np.array([1000.0, 1000.0, 1000.0, 1000.0])
#
#     def get_action(self, obs: dict[str, Vector4 | Vector6]):
#         """
#         Get env action with MPC controller.
#         :param obs: current env observation for goal conditioned tasks
#         :return: action vector
#         """
#
#         # If desired goal changed, we need to set a new target for the MPC
#         if np.all(self.desired_goal != obs['desired_goal']):
#             self._set_new_target(obs)
#             self.controller.x0 = obs['mpc_state'].reshape(-1, 1)
#             self.simulator.x0 = obs['mpc_state'].reshape(-1, 1)
#             self.controller.set_initial_guess()
#
#         # Get current car and target pose
#         car_pose = state6_to_pose3(obs['observation'])
#         target_pose = [self.rs_xs[self.rs_idx], self.rs_ys[self.rs_idx], self.rs_yaws[self.rs_idx]]
#
#         # Check distance between car and target positions
#         distance = euclidean(car_pose[0:2], target_pose[0:2])
#
#         # If distance is small increase index, changing next target pose (tvp function depends on index) for the MPC
#         if distance < self.cfg.rrt.rs.tolerance:
#
#             # Check to go over the path length
#             if self.rs_idx < (len(self.rs_xs) - 1):
#                 self.rs_idx += 1
#
#         # Make state column vector
#         state = obs['mpc_state'].reshape(-1, 1)
#
#         # Get control
#         u = self.controller.make_step(state).flatten()
#         self.simulator.make_step(u.reshape(-1, 1))
#
#         # Now we need to convert to [-1, 1] range
#         v_ref = scale_to_range(u[0], [-self.cfg.env.car.v_max, +self.cfg.env.car.v_max], [-1, 1])
#         phi_ref = scale_to_range(u[1], [-self.cfg.env.car.phi_max, +self.cfg.env.car.phi_max], [-1, 1])
#
#         return np.array([v_ref, phi_ref])
#
#     def get_distance(self, node_from: Node, node_to: Node):
#         """
#         Calculate distance between nodes using RS path's length.
#         :param node_from: start node
#         :param node_to: final node
#         :return: distance
#         """
#
#         # Get poses
#         pose_init = node_from.pose
#         pose_final = node_to.pose
#
#         # Get RS path
#         curvature = self.cfg.rrt.rs.curvature
#         step_size = self.cfg.rrt.rs.step_size
#         xs, ys, yaws, modes, lengths = reeds_shepp_path_planning(*pose_init, *pose_final, curvature, step_size)
#
#         # Handle cases where could not calculate route
#         if xs is None:
#             lengths = [np.inf]
#
#         return np.sum(np.abs(lengths))
#
#     def _set_new_target(self, obs: dict):
#         """
#         Function to be called when setting a new desired goal: sets RS paths and function to update MPC cost.
#         :param obs: gym env observation
#         """
#
#         # Update desired goal so that this set up is not called again
#         self.desired_goal = obs['desired_goal']
#
#         # Get current car and target pose to calculate waypoints
#         car_pose = state6_to_pose3(obs['observation'])
#         target_pose = state4_to_pose3(obs['desired_goal'])
#
#         # Get list of waypoints
#         curvature = self.cfg.rrt.rs.curvature
#         step_size = self.cfg.rrt.rs.step_size
#         xs, ys, yaws, modes, lengths = reeds_shepp_path_planning(*car_pose, *target_pose, curvature, step_size)
#         self.rs_xs = xs[1:]
#         self.rs_ys = ys[1:]
#         self.rs_yaws = yaws[1:]
#
#         # Initialize index
#         self.rs_idx = 0
#
#         # Get constants from config
#         axis_dist = self.cfg.env.car.axis_dist
#         n_horizon = self.cfg.rrt.mpc.n_horizon
#         n_robust = self.cfg.rrt.mpc.n_robust
#         step_size = self.cfg.rrt.mpc.step_size
#         v_max = self.cfg.env.car.v_max
#         v_min = -v_max
#         phi_max = self.cfg.env.car.phi_max
#         phi_min = -phi_max
#         control_cost = self.cfg.rrt.mpc.control_cost
#
#         # Initialize dynamics model
#         model_type = 'continuous'  # either 'discrete' or 'continuous'
#         model = Model(model_type)
#
#         # Set state variables
#         model.set_variable('_x', 'x')
#         model.set_variable('_x', 'y')
#         theta = model.set_variable('_x', 'theta')
#         v = model.set_variable('_x', 'v')
#         phi = model.set_variable('_x', 'phi')
#
#         # Set control variables
#         u_v = model.set_variable(var_type='_u', var_name='u_v')
#         u_phi = model.set_variable(var_type='_u', var_name='u_phi')
#
#         # Time varying parameters
#         model.set_variable(var_type='_tvp', var_name='rs_x')
#         model.set_variable(var_type='_tvp', var_name='rs_y')
#         model.set_variable(var_type='_tvp', var_name='rs_yaw')
#
#         # Set equations
#         tau_v = 0.05
#         k_v = 1.02
#         tau_phi = 0.05
#         k_phi = 1.0
#         model.set_rhs('x', v * cos(theta))
#         model.set_rhs('y', v * sin(theta))
#         model.set_rhs('theta', v * tan(phi) / axis_dist)
#         model.set_rhs('v', v*-(1/tau_v) + u_v*(k_v/tau_v))
#         model.set_rhs('phi', phi*-(1/tau_phi) + u_phi*(k_phi/tau_phi))
#
#         # Setup
#         model.setup()
#
#         # Initialize optimizer
#         mpc = MPC(model)
#         mpc.set_param(n_horizon=n_horizon, t_step=step_size, n_robust=n_robust, store_full_solution=True,
#                       nlpsol_opts={'ipopt.print_level': 0, 'ipopt.sb': 'yes', 'print_time': 0})
#
#         # Set state variables bound
#         mpc.bounds['lower', '_x', 'theta'] = -2 * np.pi
#         mpc.bounds['upper', '_x', 'theta'] = +2 * np.pi
#
#         # Set control variables bound
#         mpc.bounds['lower', '_u', 'u_v'] = v_min
#         mpc.bounds['upper', '_u', 'u_v'] = v_max
#         mpc.bounds['lower', '_u', 'u_phi'] = phi_min
#         mpc.bounds['upper', '_u', 'u_phi'] = phi_max
#
#         # Set cost for control input (avoid oscillations)
#         mpc.set_rterm(u_v=control_cost, u_phi=control_cost)
#
#         # Get template and define function
#         self.tvp_template = mpc.get_tvp_template()
#
#         def tvp_fun(t_now):
#             if len(self.rs_xs[self.rs_idx:]) <= n_horizon:
#                 for k in range(n_horizon + 1):
#                     self.tvp_template['_tvp', k, 'rs_x'] = self.rs_xs[-1]
#                     self.tvp_template['_tvp', k, 'rs_y'] = self.rs_ys[-1]
#                     self.tvp_template['_tvp', k, 'rs_yaw'] = self.rs_yaws[-1]
#             else:
#                 for k in range(n_horizon + 1):
#                     self.tvp_template['_tvp', k, 'rs_x'] = self.rs_xs[self.rs_idx + k]
#                     self.tvp_template['_tvp', k, 'rs_y'] = self.rs_ys[self.rs_idx + k]
#                     self.tvp_template['_tvp', k, 'rs_yaw'] = self.rs_yaws[self.rs_idx + k]
#             return self.tvp_template
#
#         # Set tvp fun
#         mpc.set_tvp_fun(tvp_fun)
#
#         # Finally, we need to add the m and l terms to cost function, but we do it when settings the target pose
#         self.controller = mpc
#
#         # Retrieve symbolic variables from model
#         x = self.controller.model._x['x']
#         y = self.controller.model._x['y']
#         theta = self.controller.model._x['theta']
#         x_target = self.controller.model.tvp['rs_x']
#         y_target = self.controller.model.tvp['rs_y']
#         theta_target = self.controller.model.tvp['rs_yaw']
#
#         # Set cost for 'm' and 'l' term
#         wp = self.cfg.rrt.mpc.weight_pos
#         wa = self.cfg.rrt.mpc.weight_angle
#         mterm = wp * (x - x_target) ** 2 + wp * (y - y_target) ** 2 + wa * (theta - theta_target) ** 2
#         lterm = wp * (x - x_target) ** 2 + wp * (y - y_target) ** 2 + wa * (theta - theta_target) ** 2
#         self.controller.set_objective(mterm=mterm, lterm=lterm)
#
#         # Call setup to finalize
#         self.controller.setup()
#
#         # Set simulator
#         self.simulator = Simulator(model)
#         self.simulator.set_param(t_step=step_size)
#
#         # TVP
#         tvp_sim = self.simulator.get_tvp_template()
#
#         def tvp_sim_fun(t_now):
#             tvp_sim['rs_x'] = self.rs_xs[self.rs_idx]
#             tvp_sim['rs_y'] = self.rs_ys[self.rs_idx]
#             tvp_sim['rs_yaw'] = self.rs_yaws[self.rs_idx]
#             return tvp_sim
#
#         self.simulator.set_tvp_fun(tvp_sim_fun)
#         self.simulator.setup()
#
#     def _set_rs_path(self, obs: dict):
#         """
#         Set new RS path.
#         :param obs: env observation
#         """
#
#         # Update desired goal so that this set up is not called again
#         self.desired_goal = obs['desired_goal']
#
#         # Get current car and target pose to calculate waypoints
#         car_pose = state6_to_pose3(obs['observation'])
#         target_pose = state4_to_pose3(obs['desired_goal'])
#
#         # Get list of waypoints
#         curvature = self.cfg.rrt.rs.curvature
#         step_size = self.cfg.rrt.rs.step_size
#         xs, ys, yaws, modes, lengths = reeds_shepp_path_planning(*car_pose, *target_pose, curvature, step_size)
#         self.rs_xs = xs[1:]
#         self.rs_ys = ys[1:]
#         self.rs_yaws = yaws[1:]
#
#         # Initialize index
#         self.rs_idx = 0



import pickle

import matplotlib.pyplot as plt
import numpy as np
from gekko import GEKKO
from scipy import signal

from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.configs import *
from rrt_ml.utilities.misc import *


cfg = MasterConfig()
cfg.env.general.stats = True
cfg.env.general.gui = True
cfg.env.general.max_timestep = 2000
cfg.env.car.f_val = 240
cfg.env.car.pose = [0, 0, 0]
env = CarNavigationBulletEnv(cfg)
env.reset_master()
obs = env.reset()

done = False
u = []
while not done:
    action = env.action_space.sample()
    for _ in range(np.random.randint(20, 80, 1)[0]):
        obs, reward, done, info = env.step(action)
        if done:
            break

# Get curves
u_v = remove_trailing_zeros(env.stats.mdp.actions[0, :, 0])
u_phi = remove_trailing_zeros(env.stats.mdp.actions[0, :, 1])
time = remove_trailing_zeros(env.stats.bullet.time[0, :, 0])
phi1 = remove_trailing_zeros(env.stats.bullet.car_steers[0, :, 0])
phi2 = remove_trailing_zeros(env.stats.bullet.car_steers[0, :, 1])
vs = remove_trailing_zeros(env.stats.bullet.car_velocities[0, :, 0])

m = GEKKO()
dt = 1/240
m.time = np.arange(0, 2000*dt-dt, dt)

# Parameters
v_meas = m.Param(value=vs)
uv_meas = m.Param(value=u_v)
v_pred = m.Var()
kvx = m.FV(value=1)
# kphix = m.FV(value=1)
kvu = m.FV(value=1)
# kphiu = m.FV(value=1)

# Available to optimize
kvx.STATUS = 1
# kphix.STATUS = 1
kvu.STATUS = 1
# kphiu.STATUS = 1

# ODE's
m.Equation(v_pred.dt() == kvx*v_pred + kvu*uv_meas)

# Objective
m.Minimize((v_pred-v_meas)**2)

# Application options
m.options.IMODE = 2   # Dynamic Simultaneous - estimation

# Solve
m.solve(disp=True)

# show final objective
print('Final SSE Objective: ' + str(m.options.objfcnval))

# Test
A = [kvx.value[0]]
B = [kvu.value[0]]
C = [1.0]
D = [0.0]
sys = signal.StateSpace(A, B, C, D)
t_model, y_model = signal.step(sys)

# Env
env.reset_master()
env.reset()
for _ in range(25):
    obs, reward, done, info = env.step([1, 0])

vs = remove_trailing_zeros(env.stats.bullet.car_velocities[0, :, 0])

fig, ax = plt.subplots()
ax.plot(np.arange(0, (1/240)*25, (1/240)), vs)
ax.plot(t_model, y_model, '--')
plt.show()



import numpy as np

from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.configs import *


cfg = MasterConfig()
cfg.env.general.gui = False
cfg.env.car.f_val = 240
cfg.env.general.max_timestep = 2000
env = CarNavigationBulletEnv(cfg)

done = False
obs = env.reset()
i = 0
while not done:
    obs, reward, done, info = env.step(env.action_space.sample())





import mrl
import torch
from typing import Callable
import os
import pickle
import dill

class PytorchModel(mrl.Module):
  """
  Generic wrapper for a pytorch nn.Module (e.g., the actorcritic network).
  These live outside of the learning algorithm modules so that they can easily be 
  shared by different modules (e.g., critic can be used by intrinsic curiosity module). 
  They are also saved independently of the agent module (which is stateless). 
  """

  def __init__(self, name : str, model_fn : Callable):
    super().__init__(name, required_agent_modules=[], locals=locals())
    self.model_fn = model_fn
    self.model = self.model_fn()

  def _setup(self):
    if self.config.get('device'):
      self.model = self.model.to(self.config.device)

  def save(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    torch.save(self.model.state_dict(), path)

  def load(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    self.model.load_state_dict(torch.load(path), strict=False)

  def copy(self, new_name):
    """Makes a copy of the Model; e.g., for target networks"""
    new_model = dill.loads(dill.dumps(self.model))
    model_fn = lambda: new_model
    return self.__class__(new_name, model_fn)

  def __call__(self, *args, **kwargs):
    if self.training:
      self.model.train()
    else:
      self.model.eval()
    return self.model(*args, **kwargs)


import torch
import numpy as np
from mrl.utils.misc import AnnotatedAttrDict

default_ddpg_config = lambda: AnnotatedAttrDict(
    device=('cuda' if torch.cuda.is_available() else 'cpu', 'torch device (cpu or gpu)'),
    gamma=(0.99, 'discount factor'),
    actor_lr=(1e-3, 'actor learning rate'),
    critic_lr=(1e-3, 'critic learning rate'),
    actor_weight_decay=(0., 'weight decay to apply to actor'),
    action_l2_regularization=(1e-2, 'l2 penalty for action norm'),
    critic_weight_decay=(0., 'weight decay to apply to critic'),
    optimize_every=(2, 'how often optimize is called, in terms of environment steps'),
    batch_size=(2000, 'batch size for training the actors/critics'),
    warm_up=(10000, 'minimum steps in replay buffer needed to optimize'),  
    initial_explore=(10000, 'steps that actor acts randomly for at beginning of training'), 
    grad_norm_clipping=(-1., 'gradient norm clipping'),
    grad_value_clipping=(-1., 'gradient value clipping'),
    target_network_update_frac=(0.005, 'polyak averaging coefficient for target networks'),
    target_network_update_freq=(1, 'how often to update target networks; NOTE: TD3 uses this too!'),
    clip_target_range=((-np.inf, np.inf), 'q/value targets are clipped to this range'),
    td3_noise=(0.1, 'noise added to next step actions in td3'),
    td3_noise_clip=(0.3, 'amount to which next step noise in td3 is clipped'),
    td3_delay=(2, 'how often the actor is trained, in terms of critic training steps, in td3'),
    entropy_coef=(0.2, 'Entropy regularization coefficient for SAC'),
    policy_opt_noise=(0., 'how much policy noise to add to actor optimization'),
    action_noise=(0.1, 'maximum std of action noise'),
    eexplore=(0., 'how often to do completely random exploration (overrides action noise)'),
    go_eexplore=(0.1, 'epsilon exploration bonus from each point of go explore, when using intrinsic curiosity'),
    go_reset_percent=(0., 'probability to reset episode early for each point of go explore, when using intrinsic curiosity'),
    overshoot_goal_percent=(0., 'if using instrinsic FIRST VISIT goals, should goal be overshot on success?'),
    direct_overshoots=(False, 'if using overshooting, should it be directed in a straight line?'),
    dg_score_multiplier=(1., 'if using instrinsic goals, score multiplier for goal candidates that are in DG distribution'),
    cutoff_success_threshold=(0.3, 0.7), # thresholds for decreasing/increasing the cutoff
    initial_cutoff=(-3, 'initial (and minimum) cutoff for intrinsic goal curiosity'),
    activ=('gelu', 'activation to use for hidden layers in networks'),
    curiosity_beta=(-3., 'beta to use for curiosity_alpha module'),
    sigma_l2_regularization=(0., 'l2 regularization on sigma critics log variance'),

    # Below are args to other modules (maybe should live in those modules?)
    seed=(0, 'random seed'),
    replay_size=(int(1e6), 'maximum size of replay buffer'),
    save_replay_buf=(False, 'save replay buffer checkpoint during training?'),
    num_envs=(12, 'number of parallel envs to run'),
    num_eval_envs=(1, 'number of parallel eval envs to run'),
    log_every=(5000, 'how often to log things'),
    varied_action_noise=(False, 'if true, action noise for each env in vecenv is interpolated between 0 and action noise'),
    use_actor_target=(False, 'if true, use actor target network to act in the environment'),
    her=('futureactual_2_2', 'strategy to use for hindsight experience replay'),
    prioritized_mode=('none', 'buffer prioritization strategy'),
    future_warm_up=(25000, 'minimum steps in replay buffer needed to stop doing ONLY future sampling'),  
    sparse_reward_shaping=(0., 'coefficient of euclidean distance reward shaping in sparse goal envs'),
    n_step_returns=(1, 'if using n-step returns, how many steps?'),
    slot_based_state=(False, 'if state is organized by slot; i.e., [batch_size, num_slots, slot_feats]')
)

def protoge_config():
  config = default_ddpg_config()
  config.gamma = 0.98
  config.actor_lr = 1e-3
  config.critic_lr = 1e-3
  config.actor_weight_decay = 0.
  config.action_l2_regularization = 1e-1
  config.target_network_update_freq = 40
  config.target_network_update_frac = 0.05
  config.optimize_every = 1
  config.batch_size = 2000
  config.warm_up = 2500
  config.initial_explore = 5000
  config.replay_size = int(1e6)
  config.clip_target_range = (-50.,0.)
  config.action_noise = 0.1
  config.eexplore = 0.1
  config.go_eexplore = 0.1
  config.go_reset_percent = 0.
  config.her = 'rfaab_1_4_3_1_1'
  config.grad_value_clipping = 5.
  return config

def best_slide_config():
  config = protoge_config()
  config.batch_size = 1000
  config.eexplore = 0.2
  config.action_noise = 0.1
  config.grad_value_clipping = -1
  config.her = 'futureactual_2_2'
  config.replay_size = int(2.5e6)
  config.initial_explore = 10000
  config.warm_up = 5000
  config.action_l2_regularization = 1e-2
  config.optimize_every = 2
  config.target_network_update_freq = 10
  config.activ = 'relu'
  return config

def protoge_td3_config():
  config = default_ddpg_config()
  config.gamma = 0.99
  config.actor_lr = 1e-3
  config.critic_lr = 1e-3
  config.actor_weight_decay = 0.
  config.target_network_update_freq = 40
  config.target_network_update_frac = 0.05
  config.optimize_every = 2
  config.batch_size = 1000
  config.warm_up = 2500
  config.replay_size = int(1e6)
  config.action_noise = 0.1
  config.eexplore = 0.1
  config.grad_value_clipping = 5.
  return config

def spinning_up_td3_config():
  config = default_ddpg_config()
  config.gamma = 0.99
  config.replay_size = int(1e6)
  config.target_network_update_frac = 0.005
  config.target_network_update_freq = 2
  config.actor_lr = 1e-3
  config.critic_lr = 1e-3
  config.action_noise = 0.1
  config.td3_noise = 0.2
  config.td3_noise_clip = 0.5
  config.td3_delay = 2
  config.batch_size = 100
  config.warm_up = 1000
  config.optimize_every = 1
  config.action_l2_regularization = 0
  config.activ = 'relu'

  # hidden sizes = (400, 300)
  # no layer norm

  return config

def spinning_up_sac_config():
  config = default_ddpg_config()
  config.gamma = 0.99
  config.replay_size = int(1e6)
  config.target_network_update_frac = 0.005
  config.target_network_update_freq = 1
  config.actor_lr = 1e-3
  config.critic_lr = 1e-3
  config.action_noise = 0.1
  config.td3_noise = 0.2
  config.td3_noise_clip = 0.5
  config.td3_delay = 2
  config.batch_size = 100
  config.warm_up = 1000
  config.optimize_every = 1
  config.action_l2_regularization = 0
  config.activ = 'relu'

  # hidden sizes = (256, 256)
  # no layer norm

  return config

def spinning_up_ddpg_config():
  config = spinning_up_td3_config()
  config.target_network_update_freq = 1
  config.activ = 'relu'
  return config

def td3_config():
  config = spinning_up_td3_config()
  config.actor_lr = 3e-4
  config.critic_lr = 3e-4
  config.batch_size = 256

  # hidden sizes = (256, 256)
  # no layer norm

  return config




from typing import Any

import numpy as np

from rrt_ml.utilities.formulas import *
from rrt_ml.utilities.hints import *
from rrt_ml.utilities.stats import *


class RRTEpochInfo(BaseStats):
    train_losses: list[float, ...] | None = None
    val_losses: list[float, ...] | None = None

    @classmethod
    def new(cls, train_losses, val_losses):
        epoch_info = cls()
        epoch_info.train_losses = train_losses
        epoch_info.val_losses = val_losses
        return epoch_info


class SLEpochInfo(BaseStats):
    epoch_n: int | None = None
    train_loss: Vector | None = None
    val_loss: Vector | None = None
    train_kl_loss: Vector | None = None
    val_kl_loss: Vector | None = None
    train_recon_loss: Vector | None = None
    val_recon_loss: Vector | None = None

    @classmethod
    def new(cls, epoch_n, t_l, v_l, t_kl_l, v_kl_l, t_r_l, v_r_l):
        self = cls()
        self.epoch_n = epoch_n
        self.train_loss = t_l
        self.val_loss = v_l
        self.train_kl_loss = t_kl_l
        self.val_kl_loss = v_kl_l
        self.train_recon_loss = t_r_l
        self.val_recon_loss = v_r_l
        return self


class RLEpochInfo(BaseStats):
    det_val_reward: float | None = None
    rand_val_reward: float | None = None
    env_stats: EnvStats | None = None
    train_timestep_num: int | None = None

    @classmethod
    def new(cls, det_val_reward, rand_val_reward, env_stats, train_timestep_num):
        self = cls()
        self.det_val_reward = det_val_reward
        self.rand_val_reward = rand_val_reward
        self.env_stats = env_stats
        self.train_timestep_num = train_timestep_num
        return self


class NodeReachInfo(BaseStats):
    cost: None | float | int = None
    done_info: None | dict = None
    length: None | float = None
    rewards: None | np.ndarray = None
    steers1: None | np.ndarray = None
    steers2: None | np.ndarray = None
    timesteps: None | np.ndarray = None
    vs: None | np.ndarray = None
    xs: None | np.ndarray = None
    yaws: None | np.ndarray = None
    ys: None | np.ndarray = None

    @classmethod
    def new_from_bullet(cls, timesteps, done_info, env_stats: EnvStats):

        # Instantiate and assign args
        stats = cls()
        stats.done_info = done_info
        stats.timesteps = timesteps

        # Secondary stats
        stats.rewards = np.trim_zeros(env_stats.bullet.rewards[0, :, 0])
        stats.steers1 = np.trim_zeros(env_stats.bullet.car_steers[0, :, 0])
        stats.steers2 = np.trim_zeros(env_stats.bullet.car_steers[0, :, 1])
        stats.vs = np.trim_zeros(env_stats.bullet.car_velocities[0, :, 0])
        stats.xs = np.trim_zeros(env_stats.bullet.car_poses[0, :, 0])
        stats.yaws = np.trim_zeros(env_stats.bullet.car_poses[0, :, 2])
        stats.ys = np.trim_zeros(env_stats.bullet.car_poses[0, :, 1])

        # Length
        stats.length = path_length(stats.xs, stats.ys)

        # Get cost
        stats.cost = stats.timesteps

        return stats

    @classmethod
    def new_from_rs(cls, timesteps, done_info, poses, rewards):

        # Instantiate and assign args
        stats = cls()
        stats.timesteps = timesteps
        stats.done_info = done_info
        poses = np.array(poses)
        stats.xs = np.array(poses[:, 0])
        stats.yaws = np.array(poses[:, 2])
        stats.ys = np.array(poses[:, 1])
        stats.rewards = np.array(rewards)

        # Secondary stats
        stats.steers1 = None
        stats.steers2 = None
        stats.vs = None

        # Length
        stats.length = path_length(stats.xs, stats.ys)

        # Get cost
        stats.cost = stats.length

        return stats


class RRTSolutionInfo(BaseStats):
    ep_stats: None | EnvStats = None
    cpu_time: None | float = None
    list_nodes: None | list[Any, ...] = None
    success: None | bool = None

    @classmethod
    def new(cls, ep_stats, cpu_time, list_nodes, success):
        stats = cls()
        stats.ep_stats = ep_stats
        stats.cpu_time = cpu_time
        stats.list_nodes = list_nodes
        stats.success = success
        return stats





import mrl
from mrl.utils.misc import soft_update, flatten_state
from mrl.modules.model import PytorchModel

import numpy as np
import torch
import torch.nn.functional as F
import os

class QValuePolicy(mrl.Module):
  """ For acting in the environment"""
  def __init__(self):
    super().__init__(
        'policy',
        required_agent_modules=[
            'qvalue', 'env', 'replay_buffer'
        ],
        locals=locals())    
  
  def _setup(self):
    self.use_qvalue_target = self.config.get('use_qvalue_target') or False
  
  def __call__(self, state, greedy=False):
    res = None
    # Initial Exploration 
    if self.training:
      if self.config.get('initial_explore') and len(
        self.replay_buffer) < self.config.initial_explore:
        res = np.array([self.env.action_space.sample() for _ in range(self.env.num_envs)])
      elif hasattr(self, 'ag_curiosity'):
        state = self.ag_curiosity.relabel_state(state)
    
    state = flatten_state(state)  # flatten goal environments
    if hasattr(self, 'state_normalizer'):
      state = self.state_normalizer(state, update=self.training)

    if res is not None:
      return res

    state = self.torch(state)
  
    if self.use_qvalue_target:
      q_values = self.numpy(self.qvalue_target(state))
    else:
      q_values = self.numpy(self.qvalue(state))
    
    if self.training and not greedy and np.random.random() < self.config.random_action_prob(steps=self.config.env_steps): 
        action = np.random.randint(self.env.action_space.n, size=[self.env.num_envs])
    else:
      action = np.argmax(q_values, -1) # Convert to int

    return action


class BaseQLearning(mrl.Module):
  """ Generic Discrete Action Q-Learning Algorithm"""

  def __init__(self):
    super().__init__(
        'algorithm',
        required_agent_modules=['qvalue','replay_buffer', 'env'],
        locals=locals())
    
  def _setup(self):
    """ Set up Q-value optimizers and create target network modules."""

    self.targets_and_models = []
    
    # Q-Value setup
    qvalue_params = []
    self.qvalues = []
    for module in list(self.module_dict.values()):
      name = module.module_name
      if name.startswith('qvalue') and isinstance(module, PytorchModel):
        self.qvalues.append(module)
        qvalue_params += list(module.model.parameters())
        target = module.copy(name + '_target')
        target.model.load_state_dict(module.model.state_dict())
        self.agent.set_module(name + '_target', target)
        self.targets_and_models.append((target.model, module.model))

    self.qvalue_opt = torch.optim.Adam(
        qvalue_params,
        lr=self.config.qvalue_lr,
        weight_decay=self.config.qvalue_weight_decay)

    self.qvalue_params = qvalue_params
          
  def save(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    torch.save({
      'qvalue_opt_state_dict': self.qvalue_opt.state_dict(),
    }, path)

  def load(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    checkpoint = torch.load(path)
    self.qvalue_opt.load_state_dict(checkpoint['qvalue_opt_state_dict'])

  def _optimize(self):
    if len(self.replay_buffer) > self.config.warm_up:
      states, actions, rewards, next_states, gammas = self.replay_buffer.sample(
          self.config.batch_size)

      self.optimize_from_batch(states, actions, rewards, next_states, gammas)
      
      if self.config.opt_steps % self.config.target_network_update_freq == 0:
        for target_model, model in self.targets_and_models:
          soft_update(target_model, model, self.config.target_network_update_frac)

  def optimize_from_batch(self, states, actions, rewards, next_states, gammas):
    raise NotImplementedError('Subclass this!')

class DQN(BaseQLearning):

  def optimize_from_batch(self, states, actions, rewards, next_states, gammas):
    q_next = self.qvalue_target(next_states).detach()

    if self.config.double_q:
      best_actions = torch.argmax(self.qvalue(next_states), dim=-1, keepdims=True)
      q_next = q_next.gather(-1, best_actions)
    else:
      q_next = q_next.max(-1, keepdims=True)[0] # Assuming action dim is the last dimension

    target = (rewards + gammas * q_next)
    target = torch.clamp(target, *self.config.clip_target_range).detach()

    if hasattr(self, 'logger') and self.config.opt_steps % 1000 == 0:
      self.logger.add_histogram('Optimize/Target_q', target)
    
    q = self.qvalue(states)
    q = q.gather(-1, actions.unsqueeze(-1).to(torch.int64))
    td_loss = F.mse_loss(q, target)

    self.qvalue_opt.zero_grad()
    td_loss.backward()
      
    # Grad clipping
    if self.config.grad_norm_clipping > 0.:	
      torch.nn.utils.clip_grad_norm_(self.qvalue_params, self.config.grad_norm_clipping)
    if self.config.grad_value_clipping > 0.:
      torch.nn.utils.clip_grad_value_(self.qvalue_params, self.config.grad_value_clipping)

    self.qvalue_opt.step()

    return



"""This module collects operations on tensors used in Ensemble-PyTorch."""


import torch
import torch.nn.functional as F


__all__ = [
    "average",
    "sum_with_multiplicative",
    "onehot_encoding",
    "pseudo_residual_classification",
    "pseudo_residual_regression",
]


def average(outputs):
    """Compute the average over a list of tensors with the same size."""
    return sum(outputs) / len(outputs)


def sum_with_multiplicative(outputs, factor):
    """
    Compuate the summation on a list of tensors, and the result is multiplied
    by a multiplicative factor.
    """
    return factor * sum(outputs)


def onehot_encoding(label, n_classes):
    """Conduct one-hot encoding on a label vector."""
    label = label.view(-1)
    onehot = torch.zeros(label.size(0), n_classes).float().to(label.device)
    onehot.scatter_(1, label.view(-1, 1), 1)

    return onehot


def pseudo_residual_classification(target, output, n_classes):
    """
    Compute the pseudo residual for classification with cross-entropyloss."""
    y_onehot = onehot_encoding(target, n_classes)

    return y_onehot - F.softmax(output, dim=1)


def pseudo_residual_regression(target, output):
    """Compute the pseudo residual for regression with least square error."""
    if target.size() != output.size():
        msg = "The shape of target {} should be the same as output {}."
        raise ValueError(msg.format(target.size(), output.size()))

    return target - output



import os
import torch


def save(model, save_dir, logger):
    """Implement model serialization to the specified directory."""
    if save_dir is None:
        save_dir = "./"

    if not os.path.isdir(save_dir):
        os.mkdir(save_dir)

    # Decide the base estimator name
    if isinstance(model.base_estimator_, type):
        base_estimator_name = model.base_estimator_.__name__
    else:
        base_estimator_name = model.base_estimator_.__class__.__name__

    # {Ensemble_Model_Name}_{Base_Estimator_Name}_{n_estimators}
    filename = "{}_{}_{}_ckpt.pth".format(
        type(model).__name__,
        base_estimator_name,
        model.n_estimators,
    )

    # The real number of base estimators in some ensembles is not same as
    # `n_estimators`.
    state = {
        "n_estimators": len(model.estimators_),
        "model": model.state_dict(),
        "_criterion": model._criterion,
    }
    save_dir = os.path.join(save_dir, filename)

    logger.info("Saving the model to `{}`".format(save_dir))

    # Save
    torch.save(state, save_dir)

    return


def load(model, save_dir="./", logger=None):
    """Implement model deserialization from the specified directory."""
    if not os.path.exists(save_dir):
        raise FileExistsError("`{}` does not exist".format(save_dir))

    # Decide the base estimator name
    if isinstance(model.base_estimator_, type):
        base_estimator_name = model.base_estimator_.__name__
    else:
        base_estimator_name = model.base_estimator_.__class__.__name__

    # {Ensemble_Model_Name}_{Base_Estimator_Name}_{n_estimators}
    filename = "{}_{}_{}_ckpt.pth".format(
        type(model).__name__,
        base_estimator_name,
        model.n_estimators,
    )
    save_dir = os.path.join(save_dir, filename)

    if logger:
        logger.info("Loading the model from `{}`".format(save_dir))

    state = torch.load(save_dir)
    n_estimators = state["n_estimators"]
    model_params = state["model"]
    model._criterion = state["_criterion"]

    # Pre-allocate and load all base estimators
    for _ in range(n_estimators):
        model.estimators_.append(model._make_estimator())
    model.load_state_dict(model_params)


def split_data_target(element, device, logger=None):
    """Split elements in dataloader according to pre-defined rules."""
    if not (isinstance(element, list) or isinstance(element, tuple)):
        msg = (
            "Invalid dataloader, please check if the input dataloder is valid."
        )
        if logger:
            logger.error(msg)
        raise ValueError(msg)

    if len(element) == 2:
        # Dataloader with one input and one target
        data, target = element[0], element[1]
        return [data.to(device)], target.to(device)  # tensor -> list
    elif len(element) > 2:
        # Dataloader with multiple inputs and one target
        data, target = element[:-1], element[-1]
        data_device = [tensor.to(device) for tensor in data]
        return data_device, target.to(device)
    else:
        # Dataloader with invalid input
        msg = (
            "The input dataloader should at least contain two tensors - data"
            " and target."
        )
        if logger:
            logger.error(msg)
        raise ValueError(msg)



import numpy as np
import random
import gym
import torch
from types import LambdaType
from scipy.linalg import block_diag
try:
  import tensorflow as tf
except:
  tf = None

def set_global_seeds(seed):
  """
    set the seed for python random, tensorflow, numpy and gym spaces

    :param seed: (int) the seed
    """
  if tf is not None:
    if hasattr(tf.random, 'set_seed'):
      tf.random.set_seed(seed)
    elif hasattr(tf.compat, 'v1'):
      tf.compat.v1.set_random_seed(seed)
    else:
      tf.set_random_seed(seed)
  np.random.seed(seed)
  random.seed(seed)
  # prng was removed in latest gym version
  if hasattr(gym.spaces, 'prng'):
    gym.spaces.prng.seed(seed)


class AttrDict(dict):
  """
    Behaves like a dictionary but additionally has attribute-style access
    for both read and write.
    e.g. x["key"] and x.key are the same,
    e.g. can iterate using:  for k, v in x.items().
    Can sublcass for specific data classes; must call AttrDict's __init__().
    """
  def __init__(self, *args, **kwargs):
    dict.__init__(self, *args, **kwargs)
    self.__dict__ = self

  def copy(self):
    """
        Provides a "deep" copy of all unbroken chains of types AttrDict, but
        shallow copies otherwise, (e.g. numpy arrays are NOT copied).
        """
    return type(self)(**{k: v.copy() if isinstance(v, AttrDict) else v for k, v in self.items()})


class AnnotatedAttrDict(AttrDict):
  """
  This is an AttrDict that accepts tuples of length 2 as values, where the
  second element is an annotation.
  """
  def __init__(self, *args, **kwargs):
    argdict = dict(*args, **kwargs)
    valuedict = {}
    annotationdict = {}
    for k, va in argdict.items():
      if hasattr(va, '__len__') and len(va) == 2 and type(va[1]) == str:
        v, a = va
        valuedict[k] = v
        annotationdict[k] = a
      else:
        valuedict[k] = va
    super().__init__(self, **valuedict)
    self.annotationdict = annotationdict

  def get_annotation(self, key):
    return self.annotationdict.get(key)


def soft_update(target, src, factor):
  with torch.no_grad():
    for target_param, param in zip(target.parameters(), src.parameters()):
      target_param.data.mul_(1.0 - factor)
      target_param.data.add_(factor * param.data)
 
def short_timestamp():
  """Returns string with timestamp"""
  import datetime
  return '{:%m%d%H%M%S}'.format(datetime.datetime.now())


def flatten_state(state):
  if isinstance(state, dict):
    obs = state['observation']
    goal = state['desired_goal']
    return np.concatenate((obs, goal), -1)
  return state


def add_config_args(argparser, config: AnnotatedAttrDict):
  """TODO: Make this add more types of args automatically?  """
  for k, v in config.items():
    try:
      if type(v) in (str, int, float):
        argparser.add_argument('--' + k, default=v, type=type(v), help=config.get_annotation(k))
      elif type(v) == bool:
        argparser.add_argument('--' + k, default=v, type=str2bool, help=config.get_annotation(k))
    except:
      pass
  return argparser


def str2bool(v):
  if isinstance(v, bool):
    return v
  if v.lower() in ('yes', 'true', 't', 'y', '1'):
    return True
  elif v.lower() in ('no', 'false', 'f', 'n', '0'):
    return False
  else:
    raise argparse.ArgumentTypeError('Boolean value expected.')


def merge_args_into_config(args, config: AttrDict):
  config.parent_folder = args.parent_folder

  other_args = {}
  for k, v in args.__dict__.items():
    if k in config:
      config[k] = v
    elif not isinstance(v, LambdaType):
      other_args[k] = v

  config.other_args = other_args

  return config


def make_agent_name(config, attr_list, prefix='agent'):
  agent_name = prefix
  attr_set = set()
  for attr in attr_list:
    s = shorten_attr(attr, attr_set)
    attr_set.add(s)
    if attr in config:
      agent_name += '_' + s + str(config[attr])
    elif attr in config.other_args:
      agent_name += '_' + s + '-' + str(config.other_args[attr])
    else:
      raise ValueError('Attribute {} not found in config!'.format(attr))
  return agent_name


def shorten_attr(attr, set, proposed_len=5):
  short = attr[:proposed_len]
  if short in set:
    return shorten_attr(attr, set, proposed_len + 1)
  return short


def softmax(X, theta=1.0, axis=None):
  """
    Compute the softmax of each element along an axis of X.

    Parameters
    ----------
    X: ND-Array. Probably should be floats.
    theta (optional): float parameter, used as a multiplier
        prior to exponentiation. Default = 1.0
    axis (optional): axis to compute values along. Default is the
        first non-singleton axis.

    Returns an array the same size as X. The result will sum to 1
    along the specified axis.
    """

  # make X at least 2d
  y = np.atleast_2d(X)

  # find axis
  if axis is None:
    axis = next(j[0] for j in enumerate(y.shape) if j[1] > 1)

  # multiply y against the theta parameter,
  y = y * float(theta)

  # subtract the max for numerical stability
  y = y - np.max(y, axis=axis, keepdims=True)

  # exponentiate y
  y = np.exp(y)

  # take the sum along the specified axis
  ax_sum = np.sum(y, axis=axis, keepdims=True)

  # finally: divide elementwise
  p = y / ax_sum

  # flatten if X was 1D
  if len(X.shape) == 1: p = p.flatten()

  return p

def make_activ(activ_name):
  if activ_name.lower() == 'relu':
    return torch.nn.ReLU
  elif activ_name.lower() == 'gelu':
    from mrl.utils.networks import GELU
    return GELU
  elif activ_name.lower() == 'tanh':
    return torch.nn.Tanh
  else:
    raise NotImplementedError


def batch_block_diag(a, b):
  """ 
  This does what scipy.linalg.block_diag does but in batch mode and with only 2 array
  https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.linalg.block_diag.html
  """
  a_shape = a.shape
  b_shape = b.shape

  if len(a_shape) == 2:
    return block_diag(a, b)
  
  assert len(a_shape) == 3
  assert len(b_shape) == 3

  assert a_shape[0] == b_shape[0] # the batch dimension

  res = np.zeros((a_shape[0], a_shape[1] + b_shape[1], a_shape[2] + b_shape[2]))
  res[:,:a_shape[1], :a_shape[2]] = a
  res[:,a_shape[1]:, a_shape[2]:] = b

  return res

def batch_block_diag_many(*arrs):
  shapes = np.array([a.shape for a in arrs], dtype=np.int64) 

  if len(shapes[0]) == 2:
    return block_diag(*arrs)
  
  # shapes is now 2D: num_arrs x 3

  res = np.zeros( (shapes[0][0], shapes[:, 1].sum(), shapes[:,2].sum()) )

  r, c = 0, 0
  for i, (batch, rr, cc) in enumerate(shapes):
      res[:, r:r + rr, c:c + cc] = arrs[i]
      r += rr
      c += cc
  
  return res


import copy

import matplotlib.pyplot as plt
import numpy as np
import pybullet as p
from scipy import signal
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit, minimize

from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.configs import *
from rrt_ml.utilities.formulas import *
from rrt_ml.utilities.misc import *

cfg = MasterConfig()
cfg.env.general.stats = True
cfg.env.general.gui = True
cfg.env.general.max_timestep = 200
cfg.env.car.f_val = 240
cfg.env.car.pose = [0, 0, 0]
env = CarNavigationBulletEnv(cfg)
env.reset_master()
obs = env.reset()

done = False
u = []
while not done:
    if env.timestep < 100:
        u.append(1)
        obs, reward, done, info = env.step([1, 0])
    else:
        u.append(-1)
        obs, reward, done, info = env.step([-1, 0])

# Get curves
time = remove_trailing_zeros(env.stats.bullet.time[0, :, 0])
vs = remove_trailing_zeros(env.stats.bullet.car_velocities[0, :, 0])

# State space model
k = 1.02
tau = 0.05
A = -1.0/tau
B = k / tau
C = 1.0
D = 0.0
sys = signal.StateSpace(A, B, C, D)

tout, yout, xout = signal.lsim(sys, u, time)

fig, axs = plt.subplots(nrows=1, ncols=1, squeeze=True)
axs.scatter(time, vs)
axs.plot(time, yout, color='red')
fig.show()

env.reset_master()
obs = env.reset()
done = False
u = []
while not done:
    if env.timestep < 100:
        u.append(0.7)
        obs, reward, done, info = env.step([0, 1])
    else:
        u.append(-0.7)
        obs, reward, done, info = env.step([0, -1])

time = remove_trailing_zeros(env.stats.bullet.time[0, :, 0])
phi1 = remove_trailing_zeros(env.stats.bullet.car_steers[0, :, 0])
phi2 = remove_trailing_zeros(env.stats.bullet.car_steers[0, :, 1])

vrl, vrr, vfl, vfr, phil, phir = get_ackermann_v_rf_lr_phi_lr(v_ref=0, phi_ref=cfg.env.car.phi_max, cfg=cfg)

a = cfg.env.car.axis_dist
b = cfg.env.car.wheel_dist

phi_ref = np.arctan((2*a*np.tan(phir))/(2*a-b*np.tan(phir)))

phi_refs = []
for p in phi2:
    phi_ref = np.arctan((2 * a * np.tan(p)) / (2 * a - b * np.tan(p)))
    phi_refs.append(phi_ref)

# State space model
k = 1.0
tau = 0.05
A = -1.0/tau
B = k / tau
C = 1.0
D = 0.0
sys = signal.StateSpace(A, B, C, D)

tout, yout, xout = signal.lsim(sys, u, time)

fig, axs = plt.subplots(nrows=1, ncols=1, squeeze=True)
axs.scatter(time, phi1)
axs.scatter(time, phi2)
axs.scatter(time, phi_refs)
axs.plot(time, yout, color='red')
fig.show()

# Second order approximation
# env.reset_master()
# obs = env.reset()
# done = False
# u = []
# for _ in range(100):
#     if env.timestep < 150:
#         u.append(1)
#         obs, reward, done, info = env.step([1, 0])
#     else:
#         u.append(-1)
#         obs, reward, done, info = env.step([-1, 0])
#
# # Get curves
# time = remove_trailing_zeros(env.stats.bullet.time[0, :, 0])
# vs = remove_trailing_zeros(env.stats.bullet.car_velocities[0, :, 0])
#
# # State space model
# k = 1.02
# tau1 = 0.02559706
# tau2 = 0.02559741
# A = [[-1/tau1, 0], [1/tau2, -1/tau2]]
# B = [[k/tau1], [0]]
# C = [0, 1]
# D = 0
# sys = signal.StateSpace(A, B, C, D)
# tout, yout, xout = signal.lsim(sys, u, time)
#
# fig, axs = plt.subplots(nrows=1, ncols=1, squeeze=True)
# axs.scatter(time, vs)
# axs.plot(time, yout, color='red')
# fig.show()

def evaluate(x):
    k = x[0]
    tau1 = x[1]
    tau2 = x[2]
    A = [[-1 / tau1, 0], [1 / tau2, -1 / tau2]]
    B = [[k / tau1], [0]]
    C = [0, 1]
    D = 0
    sys = signal.StateSpace(A, B, C, D)
    tout, yout, xout = signal.lsim(sys, u, time)
    return np.mean(np.power(yout-vs, 2))

x0 = np.array([1.02, 0.1, 0.01])
res = minimize(evaluate, x0, method='BFGS')

pass
















# State space model
k = 1.0
tau = 0.05
A = -1.0/tau
B = k / tau
C = 1.0
D = 0.0
sys = signal.StateSpace(A, B, C, D)
tout, yout, xout = signal.lsim(sys, u, time)

def evaluate(x):
    k = x[0]
    tau = x[1]
    A = -1.0 / tau
    B = k / tau
    C = 1.0
    D = 0.0
    sys = signal.StateSpace(A, B, C, D)
    tout, yout, xout = signal.lsim(sys, u, time)
    return np.mean(np.power(yout-vs, 2))

x0 = np.array([1.02, 0.1, 0.01])
res = minimize(evaluate, x0, method='BFGS')

k = res.x[0]
tau = res.x[1]
A = -1.0 / tau
B = k / tau
C = 1.0
D = 0.0
sys = signal.StateSpace(A, B, C, D)
tout, yout, xout = signal.lsim(sys, u, time)

fig, axs = plt.subplots(nrows=1, ncols=1, squeeze=True)
axs.scatter(time, vs)
axs.plot(time, yout, color='red')
fig.show()


from abc import ABC
from pathlib import Path

import matplotlib
import numpy as np
import torch as t
import yaml
from joblib import load, dump
from rich.console import Console
from torch.nn import Module
from torch.optim import Adam
from torch.utils.tensorboard import SummaryWriter

from rrt_ml.utilities.configs import MasterConfig
from rrt_ml.utilities.paths import Paths
from rrt_ml.utilities.stats import BaseStats


class Algorithm(ABC):
    """
    Abstract base class.
    """

    def __init__(self, cfg: MasterConfig):

        # Base
        self.cfg = cfg
        self.console = None  # type: Console | None
        self.epoch = None
        self.load_checkpoint = None  # type: bool | None
        self.model = None  # type: None | Module
        self.model_best_loss = None  # type: None | float
        self.model_optimizer = None  # type: None | Adam
        self.model_curr_loss = None  # type: None | float
        self.path_attrs = None  # type: Path | None
        self.path_figs = None  # type: Path | None
        self.path_model = None  # type: Path | None
        self.path_model_optimizer = None  # type: Path | None
        self.path_stats = None  # type: Path | None
        self.path_tb = None  # type: Path | None
        self.save_attrs = None  # type: list[str, ...] | None
        self.stats = None  # type: BaseStats | None
        self.tb = None  # type: SummaryWriter | None
        self.training_over = None  # type: bool | None

    def train(self, *args, **kwargs):
        """
        Train algorithm / model.
        """
        
        raise NotImplementedError
    
    def test(self, *args, **kwargs):
        """
        Test algorithm / model.
        """
        
        raise NotImplementedError

    def save_checkpoint(self, *args, **kwargs):
        """
        Save algorithm every iteration using this function.
        """

        pass

    def save_stats(self, *args, **kwargs):
        """
        Save training stats using this function.
        """

        pass

    def log_console(self, *args, **kwargs):
        """
        Log iterations on the console using this function.
        """
        
        raise NotImplementedError

    def log_tensorboard(self, *args, **kwargs):
        """
        Log iterations on tensorboard using this function.
        """
        
        raise NotImplementedError

    def _setup(self, *args, **kwargs):
        """
        General _setup.
        """

        # Set paths, load checkpoint, stats, console and tensorboard
        self._setup_paths()
        self._setup_folders()
        self._setup_save_config_to_file()
        self._setup_attrs_to_save()
        self._setup_init_base_attrs()
        self._setup_init_model()
        self._setup_checkpoint()
        self._setup_stats()
        self._setup_console()
        self._setup_tensorboard()

    def _get_checkpoint_exists_indicator(self):
        """
        Check if checkpoint exists.
        :return: boolean indicating whether checkpoint exists.
        """

        return self.path_attrs.exists()

    def _set_ignore_static_warnings(self):
        """
        Ignore linting errors.
        """

        pass

    def _set_load_attributes(self):
        """
        Load algorithm attributes.
        """

        attrs_dict = load(self.path_attrs)
        for k in attrs_dict.keys():
            if k in self.save_attrs:
                setattr(self, k, attrs_dict[k])

    def _set_load_model(self):
        """
        Load model if it exists.
        """

        # Load model
        self.model.load_state_dict(t.load(self.path_model))

        # Load optimizer and move parameters to cuda
        for state in self.model_optimizer.state.values():
            for k, v in state.items():
                if isinstance(v, t.Tensor):
                    state[k] = v.cuda() if t.cuda.is_available() else state[k]

    def _set_save_attributes(self):
        """
        Save attributes to file.
        """

        # Save attributes
        save_dict = {}
        for k, v in self.__dict__.items():
            if k in self.save_attrs:
                save_dict[k] = v
        dump(save_dict, str(self.path_attrs))

    def _set_save_model(self):
        """
        Save model.
        """

        t.save(self.model.state_dict(), self.path_model)
        t.save(self.model_optimizer.state_dict(), self.path_model_optimizer)

    def _setup_attrs_to_save(self, *args, **kwargs):
        """
        Set up list of attributes to save on checkpoint.
        """

        self.save_attrs = ['epoch', 'model_best_loss', 'model_curr_loss', 'training_over']

    def _setup_checkpoint(self, *args, **kwargs):
        """
        Initialize or load checkpoint object.
        """

        pass

    def _setup_console(self):
        """
        Initialize console logger and log initial information.
        """

        # Aliases
        alg = self.cfg.general.algorithm
        name = self.cfg.general.config_name_or_prefix

        # Create object
        self.console = Console(width=150)

        # Ruler
        self.console.rule(), self.console.rule()

        # Log config algorithm and name
        self.console.print(f"[blue bold underline]Algorithm:[/][blue] {alg}[/]", justify='center')
        self.console.print(f"[magenta bold underline]Running config:[/][magenta] {name}[/]", justify='center')

        # Log loading checkpoint
        if self.load_checkpoint:
            self.console.print(f"[red bold underline]Loading checkpoint... [/]", justify='center')

        # Ruler
        self.console.rule(), self.console.rule()

    def _setup_folders(self):
        """
        Set up experiment folders.
        """

        # Get aliases
        alg = self.cfg.general.algorithm
        name = self.cfg.general.config_name_or_prefix

        try:
            Paths().exp(alg, name).mkdir()
        except FileExistsError:
            pass

        try:
            Paths().exp_checkpoint(alg, name).mkdir()
        except FileExistsError:
            pass

        try:
            Paths().exp_stats(alg, name).mkdir()
        except FileExistsError:
            pass

        try:
            Paths().exp_tensorboard(alg, name).mkdir()
        except FileExistsError:
            pass

        try:
            Paths().exp_fig(alg, name).mkdir()
        except FileExistsError:
            pass

    def _setup_init_base_attrs(self):
        """
        Initialize base attributes (epoch, model losses, etc)
        """

        self.epoch = 0
        self.model_curr_loss = np.inf
        self.model_best_loss = np.inf
        self.training_over = False

    def _setup_init_model(self):
        """
        Initialize model.
        """

        pass

    def _setup_paths(self, *args, **kwargs):
        """
        Setup paths for checkpoints, models, etc.
        """

        # Aliases
        alg = self.cfg.general.algorithm
        name = self.cfg.general.config_name_or_prefix

        # Attributes path
        self.path_attrs = Paths().exp_checkpoint(alg, name) / "attrs"

        # Stats path
        self.path_stats = Paths().exp_stats(alg, name) / "stats"

        # Figures path
        self.path_figs = Paths().exp_fig(alg, name)

        # Tensorboard path
        self.path_tb = Paths().exp_tensorboard(alg, name)

        # Model and optimizer paths
        self.path_model = Paths().exp_checkpoint(alg, name) / "model.pt"
        self.path_model_optimizer = Paths().exp_checkpoint(alg, name) / "model_optimizer.pt"

    def _setup_save_config_to_file(self):
        """
        Save config to file.
        """

        # Aliases
        alg = self.cfg.general.algorithm
        name = self.cfg.general.config_name_or_prefix

        # Save config at experiment base folder
        path = Paths().exp(alg, name) / 'config.yaml'
        if not path.exists():
            with open(str(path), 'w') as f:
                yaml.dump(self.cfg.dict(), f)

    def _setup_stats(self, *args, **kwargs):
        """
        Set up stats object.
        """

        pass

    def _setup_tensorboard(self, *args, **kwargs):
        """
        Setup tensorboard logging.
        """

        # Don't show figure window
        # matplotlib.use('Agg')

        if not self.training_over:
            self.tb = SummaryWriter(log_dir=str(self.path_tb))



import numpy as np
from collections import OrderedDict
from mrl.utils.misc import AttrDict
from multiprocessing import RawValue

class RingBuffer(object):
  """This is a collections.deque in numpy, with pre-allocated memory"""

  def __init__(self, maxlen, shape, dtype=np.float32, data=None):
    """
    A buffer object, when full restarts at the initial position

    :param maxlen: (int) the max number of numpy objects to store
    :param shape: (tuple) the shape of the numpy objects you want to store
    :param dtype: (str) the name of the type of the numpy object you want to store
    """
    self.maxlen = maxlen
    self.start = RawValue('L')
    self.length = RawValue('L')
    self.shape = shape
    if data is None:
      self.data = np.zeros((maxlen, ) + shape, dtype=dtype)
    else:
      assert data.shape == (maxlen, ) + shape
      assert data.dtype == dtype
      self.data = data

  def _get_state(self):
    # Only restore the values in the data
    end_idx = self.start.value + self.length.value
    indices = range(self.start.value, end_idx)
    return self.start.value, self.length.value, self.data.take(indices, axis=0, mode='wrap')

  def _set_state(self, start, length, data):
    self.start.value = start
    self.length.value = length
    self.data[:length] = data
    self.data = np.roll(self.data, start, axis=0)

  def __len__(self):
    return self.length.value

  def __getitem__(self, idx):
    if idx < 0 or idx >= self.length.value:
      raise KeyError()
    return self.data[(self.start.value + idx) % self.maxlen]

  def get_batch(self, idxs):
    """
    get the value at the indexes

    :param idxs: (int or numpy int) the indexes
    :return: (np.ndarray) the stored information in the buffer at the asked positions
    """
    return self.data[(self.start.value + idxs) % self.length.value]

  def append(self, var):
    """
    Append an object to the buffer

    :param var: (np.ndarray) the object you wish to add
    """
    if self.length.value < self.maxlen:
      # We have space, simply increase the length.
      self.length.value += 1
    elif self.length.value == self.maxlen:
      # No space, "remove" the first item.
      self.start.value = (self.start.value + 1) % self.maxlen
    else:
      # This should never happen.
      raise RuntimeError()

    self.data[(self.start.value + self.length.value - 1) % self.maxlen] = var

  def _append_batch_with_space(self, var):
    """
    Append a batch of objects to the buffer, *assuming* there is space.

    :param var: (np.ndarray) the batched objects you wish to add
    """
    len_batch = len(var)
    start_pos = (self.start.value + self.length.value) % self.maxlen

    self.data[start_pos : start_pos + len_batch] = var
    
    if self.length.value < self.maxlen:
      self.length.value += len_batch
      assert self.length.value <= self.maxlen, "this should never happen!"
    else:
      self.start.value = (self.start.value + len_batch) % self.maxlen

    return np.arange(start_pos, start_pos + len_batch)

  def append_batch(self, var):
    """
    Append a batch of objects to the buffer.

    :param var: (np.ndarray) the batched objects you wish to add
    """
    len_batch = len(var)
    assert len_batch < self.maxlen, 'trying to add a batch that is too big!'
    start_pos = (self.start.value + self.length.value) % self.maxlen
    
    if start_pos + len_batch <= self.maxlen:
      # If there is space, add it
      idxs = self._append_batch_with_space(var)
    else:
      # No space, so break it into two batches for which there is space
      first_batch, second_batch = np.split(var, [self.maxlen - start_pos])
      idxs1 = self._append_batch_with_space(first_batch)
      # use append on second call in case len_batch > self.maxlen
      idxs2 = self._append_batch_with_space(second_batch)
      idxs = np.concatenate((idxs1, idxs2))
    return idxs

class ReplayBuffer(object):
  def __init__(self, limit, item_shape, dtypes=None):
    """
    The replay buffer object

    :param limit: (int) the max number of transitions to store
    :param item_shape: a list of tuples of (str) item name and (tuple) the shape for item
      Ex: [("observations0", env.observation_space.shape),\
          ("actions",env.action_space.shape),\
          ("rewards", (1,)),\
          ("observations1",env.observation_space.shape ),\
          ("terminals1", (1,))]
    :param dtypes: list of dtype tuples; useful for storing things as float16.
    """
    self.limit = limit

    self.items = OrderedDict()
    if dtypes is None:
      dtypes = [(np.float32, np.float32)] * len(item_shape)

    self.in_types, self.out_types = zip(*dtypes)

    for (name, shape), dtype in zip(item_shape, self.in_types):
      self.items[name] = RingBuffer(limit, shape=shape, dtype=dtype)

  def sample(self, batch_size):
    """
    sample a random batch from the buffer

    :param batch_size: (int) the number of element to sample for the batch
    :return: (list) the sampled batch
    """
    if self.size==0:
      return []

    batch_idxs = np.random.randint(self.size, size=batch_size)

    transition = []
    for buf, dtype in zip(self.items.values(), self.out_types):
      item = buf.get_batch(batch_idxs).astype(dtype)
      transition.append(item)

    return transition

  def add(self, *items):
    """
    Appends a single transition to the buffer

    :param items: a list of values for the transition to append to the replay buffer,
        in the item order that we initialized the ReplayBuffer with.
    """
    for buf, value in zip(self.items.values(), items):
      buf.append(value)

  def add_batch(self, *items):
    """
    Append a batch of transitions to the buffer.

    :param items: a list of batched transition values to append to the replay buffer,
        in the item order that we initialized the ReplayBuffer with.
    """
    if (items[0].shape) == 1 or len(items[0]) == 1:
      self.add(*items)
      return

    for buf, batched_values in zip(self.items.values(), items):
      idxs = buf.append_batch(batched_values)

    return idxs

  def __len__(self):
    return self.size

  def _get_state(self):
    d = dict()
    for item, buf in self.items.items():
      d[item] = buf._get_state()
    return d

  def _set_state(self, d):
    for item, buf in self.items.items():
      buf._set_state(*d[item])

  @property
  def size(self):
    # Get the size of the RingBuffer on the first item type
    return len(next(iter(self.items.values())))



from pathlib import Path

import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.ticker import MaxNLocator
from scipy.interpolate import interp1d

from rrt_ml.algorithms.rl import *
from rrt_ml.utilities.configs import *


def run_rl(cfg: MasterConfig):

    # Train
    if cfg.general.is_train:

        # Initialize RL
        rl = RL(cfg)
        rl.train()

    # Test
    else:
        rl = RL(cfg)
        rl.test()

    # Test
    # elif cfg.test:

        # AgentUtils.plot_trajectory_top_view(target_pose=[-1, 0, np.pi])
        # AgentUtils.plot_success_rate()

        # Initialize RL
        # rl = RL(cfg)

        # Test episodes
        # rl.test_multiple_episodes()

        # Generate data
        # rl.get_distance_data()
        # rl.get_distance_data()

        # Test distance model
        # rl.test_distance_model()
        # rl.test_distance_model_corn()

        # Test trajectory
        # rl.test_plot_trajectory_top_view_variables(1)
        # rl.test_get_top_view_image()

        # Hyperparameter search graph
        # plot_hyperparameter_search(cfg)

        # Plot reward and entropy
        # plot_reward_entropy(cfg)


def plot_hyperparameter_search(cfg):
    """
    Results from hyperparameter search.
    :param cfg: experiment configuration.
    """

    # Experiments path
    base_path = Path(__file__).resolve().parents[2]
    exp_path = base_path / "experiments" / "rl"
    file_path = base_path.parents[0] / "latex" / "figs" / "apdx" / "rl_hiper.png"

    # Figures
    fig, axs = plt.subplots(nrows=4, ncols=2, sharex="all", sharey="all")
    fig.set_size_inches(16, 22)
    axs = axs.flatten()

    # Loop
    ax_counter = 0
    for param1 in [1]:
        for param2 in [1, 2]:
            for param3 in [1, 2]:

                # Name
                prefix = "a" + str(param1) + str(param2) + str(param3)

                # Load test rewards
                path = exp_path / prefix / (prefix + "__Test__Episode_rewards.csv")
                df = pd.read_csv(path)  # type: pd.DataFrame
                time = df["wall_time"].to_numpy()
                step = df["step"].to_numpy()
                rewards = df["Test/Episode_rewards"].to_numpy()

                # Fix arrays for interpolation
                rewards = rewards[:-1]
                step = step[:-1]

                # Load entropy
                path = exp_path / prefix / (prefix + "__Explore__ag_kde_entropy.csv")
                df = pd.read_csv(path)

                # Smoothing
                rewards = _smooth(rewards, 0.8)

                # Plot rewards
                axs[ax_counter].plot(step, rewards)
                axs[ax_counter].tick_params(axis="y", labelcolor="b")
                axs[ax_counter].yaxis.set_major_locator(
                    MaxNLocator(nbins=5, integer=True)
                )

                # Plot entropy
                ax2 = axs[ax_counter].twinx()
                x = df["step"].to_numpy()
                y = df["Explore/ag_kde_entropy"].to_numpy()
                f = interp1d(x, y)
                ax2.plot(step, f(step), color="g")
                ax2.tick_params(axis="y", labelcolor="g")
                ax2.yaxis.set_major_locator(MaxNLocator(nbins=5, integer=True))

                # Set title
                axs[ax_counter].title.set_text(
                    f"Hiperparâmetro {param1}{param2}{param3}"
                )

                # Increment counter
                ax_counter += 1

    # Save
    fig.savefig(file_path, dpi=600)


def _smooth(series, weight):
    last = series[0]  # First value in the plot (first timestep)
    smoothed = list()
    for point in series:
        smoothed_val = last * weight + (1 - weight) * point  # Calculate smoothed value
        smoothed.append(smoothed_val)  # Save it
        last = smoothed_val  # Anchor the last smoothed value
    return smoothed



import mrl
import pickle
import os
import numpy as np
import torch

class RunningMeanStd(object):
  # https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
  def __init__(self, epsilon=1e-4, shape=()):
    self.mean = np.zeros(shape, 'float64')
    self.var = np.ones(shape, 'float64')
    self.count = epsilon

  def update(self, x):
    batch_mean = np.mean(x, axis=0, keepdims=True)
    batch_var = np.var(x, axis=0, keepdims=True)
    batch_count = x.shape[0]
    self.update_from_moments(batch_mean, batch_var, batch_count)

  def update_from_moments(self, batch_mean, batch_var, batch_count):
    delta = batch_mean - self.mean
    tot_count = self.count + batch_count

    new_mean = self.mean + delta * batch_count / tot_count
    m_a = self.var * (self.count)
    m_b = batch_var * (batch_count)
    M2 = m_a + m_b + np.square(delta) * self.count * batch_count / (tot_count)
    new_var = M2 / (tot_count)

    self.mean = new_mean
    self.var = new_var
    self.count = tot_count


class Normalizer(mrl.Module):
  def __init__(self, normalizer):
    super().__init__('state_normalizer', required_agent_modules=[], locals=locals())
    self.normalizer = normalizer
    self.lazy_load = None

  def __call__(self, *args, **kwargs):
    if self.training:
      self.normalizer.read_only = False
    else:
      self.normalizer.read_only = True

    if self.lazy_load is not None:
      self.normalizer(*args, **kwargs)
      self.load(self.lazy_load)
      print("LOADED NORMALIZER")
      self.lazy_load = None

    return self.normalizer(*args, **kwargs)

  def save(self, save_folder):
    if self.normalizer.state_dict() is not None:
      with open(os.path.join(save_folder, 'normalizer.pickle'), 'wb') as f:
        pickle.dump(self.normalizer.state_dict(), f)

  def load(self, save_folder):
    if self.normalizer.state_dict() is not None:
      save_path = os.path.join(save_folder, 'normalizer.pickle')
      if os.path.exists(save_path):
        with open(save_path, 'rb') as f:
          self.normalizer.load_state_dict(pickle.load(f))
      else:
        print('WARNING: No saved normalizer state to load.')
    else:
      self.lazy_load = save_folder


# Below from https://github.com/ShangtongZhang/DeepRL/blob/master/deep_rl/utils/normalizer.py


class BaseNormalizer:
  def __init__(self, read_only=False):
    self.read_only = read_only

  def set_read_only(self):
    self.read_only = True

  def unset_read_only(self):
    self.read_only = False

  def state_dict(self):
    return None

  def load_state_dict(self, _):
    return


class MeanStdNormalizer(BaseNormalizer):
  def __init__(self, read_only=False, clip_before=200.0, clip_after=5.0, epsilon=1e-8):
    BaseNormalizer.__init__(self, read_only)
    self.read_only = read_only
    self.rms = None
    self.clip_before = clip_before
    self.clip_after = clip_after
    self.epsilon = epsilon

  def __call__(self, x, update=True):
    x = np.clip(np.asarray(x), -self.clip_before, self.clip_before)
    if self.rms is None:
      self.rms = RunningMeanStd(shape=(1, ) + x.shape[1:])
    if not self.read_only and update:
      self.rms.update(x)
    return np.clip((x - self.rms.mean) / np.sqrt(self.rms.var + self.epsilon), -self.clip_after, self.clip_after)

  def state_dict(self):
    if self.rms is not None:
      return {'mean': self.rms.mean, 'var': self.rms.var, 'count': self.rms.count}

  def load_state_dict(self, saved):
    self.rms.mean = saved['mean']
    self.rms.var = saved['var']
    self.rms.count = saved['count']


class RescaleNormalizer(BaseNormalizer):
  def __init__(self, coef=1.0):
    BaseNormalizer.__init__(self)
    self.coef = coef

  def __call__(self, x, *unused_args):
    if not isinstance(x, torch.Tensor):
      x = np.asarray(x)
    return self.coef * x


class ImageNormalizer(RescaleNormalizer):
  def __init__(self):
    RescaleNormalizer.__init__(self, 1.0 / 255)


class SignNormalizer(BaseNormalizer):
  def __call__(self, x, *unused_args):
    return np.sign(x)



"""
  In bagging-based ensemble, each base estimator is trained independently.
  In addition, sampling with replacement is conducted on the training data
  batches to encourage the diversity between different base estimators in
  the ensemble.
"""


import torch
import torch.nn as nn
import torch.nn.functional as F

import warnings
from joblib import Parallel, delayed

from ._base import BaseClassifier, BaseRegressor
from ._base import torchensemble_model_doc
from .utils import io
from .utils import set_module
from .utils import operator as op


__all__ = ["BaggingClassifier", "BaggingRegressor"]


def _parallel_fit_per_epoch(
    train_loader,
    estimator,
    cur_lr,
    optimizer,
    criterion,
    idx,
    epoch,
    log_interval,
    device,
    is_classification,
):
    """
    Private function used to fit base estimators in parallel.

    WARNING: Parallelization when fitting large base estimators may cause
    out-of-memory error.
    """
    if cur_lr:
        # Parallelization corrupts the binding between optimizer and scheduler
        set_module.update_lr(optimizer, cur_lr)

    for batch_idx, elem in enumerate(train_loader):

        data, target = io.split_data_target(elem, device)
        batch_size = data[0].size(0)

        # Sampling with replacement
        sampling_mask = torch.randint(
            high=batch_size, size=(int(batch_size),), dtype=torch.int64
        )
        sampling_mask = torch.unique(sampling_mask)  # remove duplicates
        subsample_size = sampling_mask.size(0)
        sampling_data = [tensor[sampling_mask] for tensor in data]
        sampling_target = target[sampling_mask]

        optimizer.zero_grad()
        sampling_output = estimator(*sampling_data)
        loss = criterion(sampling_output, sampling_target)
        loss.backward()
        optimizer.step()

        # Print training status
        if batch_idx % log_interval == 0:

            # Classification
            if is_classification:
                _, predicted = torch.max(sampling_output.data, 1)
                correct = (predicted == sampling_target).sum().item()

                msg = (
                    "Estimator: {:03d} | Epoch: {:03d} | Batch: {:03d}"
                    " | Loss: {:.5f} | Correct: {:d}/{:d}"
                )
                print(
                    msg.format(
                        idx, epoch, batch_idx, loss, correct, subsample_size
                    )
                )
            else:
                msg = (
                    "Estimator: {:03d} | Epoch: {:03d} | Batch: {:03d}"
                    " | Loss: {:.5f}"
                )
                print(msg.format(idx, epoch, batch_idx, loss))

    return estimator, optimizer


@torchensemble_model_doc(
    """Implementation on the BaggingClassifier.""", "model"
)
class BaggingClassifier(BaseClassifier):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in BaggingClassifier.""",
        "classifier_forward",
    )
    def forward(self, *x):
        # Average over class distributions from all base estimators.
        outputs = [
            F.softmax(estimator(*x), dim=1) for estimator in self.estimators_
        ]
        proba = op.average(outputs)

        return proba

    @torchensemble_model_doc(
        """Set the attributes on optimizer for BaggingClassifier.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for BaggingClassifier.""",
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for BaggingClassifier.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @torchensemble_model_doc(
        """Implementation on the training stage of BaggingClassifier.""", "fit"
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):

        self._validate_parameters(epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        # Instantiate a pool of base estimators, optimizers, and schedulers.
        estimators = []
        for _ in range(self.n_estimators):
            estimators.append(self._make_estimator())

        optimizers = []
        for i in range(self.n_estimators):
            optimizers.append(
                set_module.set_optimizer(
                    estimators[i], self.optimizer_name, **self.optimizer_args
                )
            )

        if self.use_scheduler_:
            scheduler_ = set_module.set_scheduler(
                optimizers[0], self.scheduler_name, **self.scheduler_args
            )

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.CrossEntropyLoss()

        # Utils
        best_acc = 0.0

        # Internal helper function on pesudo forward
        def _forward(estimators, *x):
            outputs = [
                F.softmax(estimator(*x), dim=1) for estimator in estimators
            ]
            proba = op.average(outputs)

            return proba

        # Maintain a pool of workers
        with Parallel(n_jobs=self.n_jobs) as parallel:

            # Training loop
            for epoch in range(epochs):
                self.train()

                if self.use_scheduler_:
                    cur_lr = scheduler_.get_last_lr()[0]
                else:
                    cur_lr = None

                if self.n_jobs and self.n_jobs > 1:
                    msg = "Parallelization on the training epoch: {:03d}"
                    self.logger.info(msg.format(epoch))

                rets = parallel(
                    delayed(_parallel_fit_per_epoch)(
                        train_loader,
                        estimator,
                        cur_lr,
                        optimizer,
                        self._criterion,
                        idx,
                        epoch,
                        log_interval,
                        self.device,
                        True,
                    )
                    for idx, (estimator, optimizer) in enumerate(
                        zip(estimators, optimizers)
                    )
                )

                estimators, optimizers = [], []
                for estimator, optimizer in rets:
                    estimators.append(estimator)
                    optimizers.append(optimizer)

                # Validation
                if test_loader:
                    self.eval()
                    with torch.no_grad():
                        correct = 0
                        total = 0
                        for _, elem in enumerate(test_loader):
                            data, target = io.split_data_target(
                                elem, self.device
                            )
                            output = _forward(estimators, *data)
                            _, predicted = torch.max(output.data, 1)
                            correct += (predicted == target).sum().item()
                            total += target.size(0)
                        acc = 100 * correct / total

                        if acc > best_acc:
                            best_acc = acc
                            self.estimators_ = nn.ModuleList()
                            self.estimators_.extend(estimators)
                            if save_model:
                                io.save(self, save_dir, self.logger)

                        msg = (
                            "Epoch: {:03d} | Validation Acc: {:.3f}"
                            " % | Historical Best: {:.3f} %"
                        )
                        self.logger.info(msg.format(epoch, acc, best_acc))
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "bagging/Validation_Acc", acc, epoch
                            )

                # Update the scheduler
                with warnings.catch_warnings():

                    # UserWarning raised by PyTorch is ignored because
                    # scheduler does not have a real effect on the optimizer.
                    warnings.simplefilter("ignore", UserWarning)

                    if self.use_scheduler_:
                        scheduler_.step()

        self.estimators_ = nn.ModuleList()
        self.estimators_.extend(estimators)
        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="classifier_evaluate")
    def evaluate(self, test_loader, return_loss=False):
        return super().evaluate(test_loader, return_loss)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)


@torchensemble_model_doc(
    """Implementation on the BaggingRegressor.""", "model"
)
class BaggingRegressor(BaseRegressor):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in BaggingRegressor.""",
        "regressor_forward",
    )
    def forward(self, *x):
        # Average over predictions from all base estimators.
        outputs = [estimator(*x) for estimator in self.estimators_]
        pred = op.average(outputs)

        return pred

    @torchensemble_model_doc(
        """Set the attributes on optimizer for BaggingRegressor.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for BaggingRegressor.""",
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for BaggingRegressor.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @torchensemble_model_doc(
        """Implementation on the training stage of BaggingRegressor.""", "fit"
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):

        self._validate_parameters(epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        # Instantiate a pool of base estimators, optimizers, and schedulers.
        estimators = []
        for _ in range(self.n_estimators):
            estimators.append(self._make_estimator())

        optimizers = []
        for i in range(self.n_estimators):
            optimizers.append(
                set_module.set_optimizer(
                    estimators[i], self.optimizer_name, **self.optimizer_args
                )
            )

        if self.use_scheduler_:
            scheduler_ = set_module.set_scheduler(
                optimizers[0], self.scheduler_name, **self.scheduler_args
            )

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.MSELoss()

        # Utils
        best_loss = float("inf")

        # Internal helper function on pesudo forward
        def _forward(estimators, *x):
            outputs = [estimator(*x) for estimator in estimators]
            pred = op.average(outputs)

            return pred

        # Maintain a pool of workers
        with Parallel(n_jobs=self.n_jobs) as parallel:

            # Training loop
            for epoch in range(epochs):
                self.train()

                if self.use_scheduler_:
                    cur_lr = scheduler_.get_last_lr()[0]
                else:
                    cur_lr = None

                if self.n_jobs and self.n_jobs > 1:
                    msg = "Parallelization on the training epoch: {:03d}"
                    self.logger.info(msg.format(epoch))

                rets = parallel(
                    delayed(_parallel_fit_per_epoch)(
                        train_loader,
                        estimator,
                        cur_lr,
                        optimizer,
                        self._criterion,
                        idx,
                        epoch,
                        log_interval,
                        self.device,
                        False,
                    )
                    for idx, (estimator, optimizer) in enumerate(
                        zip(estimators, optimizers)
                    )
                )

                estimators, optimizers = [], []
                for estimator, optimizer in rets:
                    estimators.append(estimator)
                    optimizers.append(optimizer)

                # Validation
                if test_loader:
                    self.eval()
                    with torch.no_grad():
                        val_loss = 0.0
                        for _, elem in enumerate(test_loader):
                            data, target = io.split_data_target(
                                elem, self.device
                            )
                            output = _forward(estimators, *data)
                            val_loss += self._criterion(output, target)
                        val_loss /= len(test_loader)

                        if val_loss < best_loss:
                            best_loss = val_loss
                            self.estimators_ = nn.ModuleList()
                            self.estimators_.extend(estimators)
                            if save_model:
                                io.save(self, save_dir, self.logger)

                        msg = (
                            "Epoch: {:03d} | Validation Loss:"
                            " {:.5f} | Historical Best: {:.5f}"
                        )
                        self.logger.info(
                            msg.format(epoch, val_loss, best_loss)
                        )
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "bagging/Validation_Loss", val_loss, epoch
                            )

                # Update the scheduler
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)

                    if self.use_scheduler_:
                        scheduler_.step()

        self.estimators_ = nn.ModuleList()
        self.estimators_.extend(estimators)
        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="regressor_evaluate")
    def evaluate(self, test_loader):
        return super().evaluate(test_loader)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)



import pandas as pd


df_train = pd.read_csv('_train.csv')
df_val = pd.read_csv('_train.csv')

del df_train['i']
del df_val['i']

df_train = df_train.rename(columns={
    'sin_theta_i': 'cos_theta_i',
    'cos_theta_i': 'sin_theta_i',
    'sin_theta_f': 'cos_theta_f',
    'cos_theta_f': 'sin_theta_f',
    'sin_theta': 'cos_theta',
    'cos_theta': 'sin_theta',
})

df_val = df_val.rename(columns={
    'sin_theta_i': 'cos_theta_i',
    'cos_theta_i': 'sin_theta_i',
    'sin_theta_f': 'cos_theta_f',
    'cos_theta_f': 'sin_theta_f',
    'sin_theta': 'cos_theta',
    'cos_theta': 'sin_theta',
})

df_train.to_csv('train.csv')
df_val.to_csv('val.csv')



# Core
import mrl
from mrl import Agent, Module, config_to_agent

# Utils
from mrl.utils.networks import *
from mrl.utils.misc import AttrDict, add_config_args, merge_args_into_config, make_agent_name, make_activ

# Replays
from mrl.replays.online_her_buffer import OnlineHERBuffer
from mrl.replays.prioritized_replay import EntropyPrioritizedOnlineHERBuffer
from mrl.replays.old_replay_buffer import OldReplayBuffer

# Configs
from mrl.configs.continuous_off_policy import *
from mrl.configs.discrete_off_policy import *

# Modules
from mrl.modules.action_noise import *
from mrl.modules.curiosity import *
from mrl.modules.density import *
from mrl.modules.env import EnvModule, FirstVisitDoneWrapper
from mrl.modules.eval import EpisodicEval
from mrl.modules.goal_reward import GoalEnvReward, NeighborReward
from mrl.modules.logging import Logger, colorize
from mrl.modules.model import PytorchModel
from mrl.modules.normalizer import *
from mrl.modules.success_prediction import GoalSuccessPredictor
from mrl.modules.train import StandardTrain

# Algorithms
from mrl.algorithms.continuous_off_policy import *
from mrl.algorithms.discrete_off_policy import QValuePolicy, DQN
from mrl.algorithms.fixed_horizon_DDPG import FHDDPG
from mrl.algorithms.random_ensemble_DPG import RandomEnsembleDPG


# Private imports if available
try:
  from private import *
except:
  print(colorize("Warning: Failed to import private modules.", color='red', bold=True))
  pass


import os
import time
import logging


_tb_logger = None
__all__ = ["set_logger", "get_tb_logger"]


def set_logger(
    log_file=None,
    log_console_level="info",
    log_file_level=None,
    use_tb_logger=False,
):
    """Bind the default logger with console and file stream output."""

    def _get_level(level):
        if level.lower() == "debug":
            return logging.DEBUG
        elif level.lower() == "info":
            return logging.INFO
        elif level.lower() == "warning":
            return logging.WARN
        elif level.lower() == "error":
            return logging.ERROR
        elif level.lower() == "critical":
            return logging.CRITICAL
        else:
            msg = (
                "`log_console_level` must be one of {{DEBUG, INFO,"
                " WARNING, ERROR, CRITICAL}}, but got {} instead."
            )
            raise ValueError(msg.format(level.upper()))

    _logger = logging.getLogger()

    # Reset
    for h in _logger.handlers:
        _logger.removeHandler(h)

    rq = time.strftime("%Y_%m_%d_%H_%M", time.localtime(time.time()))
    log_path = os.path.join(os.getcwd(), "logs")

    ch_formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s: %(message)s"
    )
    ch = logging.StreamHandler()
    ch.setLevel(_get_level(log_console_level))
    ch.setFormatter(ch_formatter)
    _logger.addHandler(ch)

    if log_file is not None:
        print("Log will be saved in '{}'.".format(log_path))
        if not os.path.exists(log_path):
            os.mkdir(log_path)
            print("Create folder 'logs/'")
        log_name = os.path.join(log_path, log_file + "-" + rq + ".log")
        print("Start logging into file {}...".format(log_name))
        fh = logging.FileHandler(log_name, mode="w")
        fh.setLevel(
            logging.DEBUG
            if log_file_level is None
            else _get_level(log_file_level)
        )
        fh_formatter = logging.Formatter(
            "%(asctime)s - %(filename)s[line:%(lineno)d] - "
            "%(levelname)s: %(message)s"
        )
        fh.setFormatter(fh_formatter)
        _logger.addHandler(fh)
    _logger.setLevel("DEBUG")

    if use_tb_logger:
        tb_log_path = os.path.join(
            log_path, log_file + "-" + rq + "_tb_logger"
        )
        os.mkdir(tb_log_path)
        init_tb_logger(log_dir=tb_log_path)

    return _logger


def init_tb_logger(log_dir):
    try:
        import tensorboard  # noqa: F401
    except ModuleNotFoundError:
        msg = (
            "Cannot load the module tensorboard. Please make sure that"
            " tensorboard is installed."
        )
        raise ModuleNotFoundError(msg)

    from torch.utils.tensorboard import SummaryWriter

    global _tb_logger

    if not _tb_logger:
        _tb_logger = SummaryWriter(log_dir=log_dir)


def get_tb_logger():
    return _tb_logger



import torch
import pytest
import numpy as np
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader

import torchensemble
from torchensemble.utils.logging import set_logger


parallel = [
    torchensemble.FusionClassifier,
    torchensemble.VotingClassifier,
    torchensemble.BaggingClassifier,
    torchensemble.SoftGradientBoostingClassifier,
]

set_logger("pytest_training_params")


# Base estimator
class MLP(nn.Module):
    def __init__(self):
        super(MLP, self).__init__()
        self.linear1 = nn.Linear(2, 2)
        self.linear2 = nn.Linear(2, 2)

    def forward(self, X):
        X = X.view(X.size()[0], -1)
        output = self.linear1(X)
        output = self.linear2(output)
        return output


# Trainining data
X_train = torch.Tensor(np.array(([1, 1], [2, 2], [3, 3], [4, 4])))

y_train = torch.LongTensor(np.array(([0, 0, 1, 1])))


# Prepare data
train = TensorDataset(X_train, y_train)
train_loader = DataLoader(train, batch_size=2)


@pytest.mark.parametrize("method", parallel)
def test_parallel(method):
    model = method(estimator=MLP, n_estimators=2, cuda=False)

    # Epochs
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, epochs=-1)
    assert "number of training epochs" in str(excinfo.value)

    # Log interval
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, log_interval=-1)
    assert "number of batches to wait" in str(excinfo.value)


def test_gradient_boosting():
    model = torchensemble.GradientBoostingClassifier(
        estimator=MLP, n_estimators=2, cuda=False
    )

    # Epochs
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, epochs=-1)
    assert "number of training epochs" in str(excinfo.value)

    # Log interval
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, log_interval=-1)
    assert "number of batches to wait" in str(excinfo.value)

    # Early stoppping round
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, early_stopping_rounds=0)
    assert "number of tolerant rounds" in str(excinfo.value)

    # Shrinkage rate
    model = torchensemble.GradientBoostingClassifier(
        estimator=MLP, n_estimators=2, shrinkage_rate=2, cuda=False
    )
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader)
    assert "shrinkage rate should be in the range" in str(excinfo.value)


def test_snapshot_ensemble():
    model = torchensemble.SnapshotEnsembleClassifier(
        estimator=MLP, n_estimators=2, cuda=False
    )

    # Learning rate clip
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, lr_clip=1)
    assert "lr_clip should be a list or tuple" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, lr_clip=[0, 1, 2])
    assert "lr_clip should only have two elements" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, lr_clip=[1, 0])
    assert "should be smaller than the second" in str(excinfo.value)

    # Epochs
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, epochs=-1)
    assert "number of training epochs" in str(excinfo.value)

    # Log interval
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, log_interval=-1)
    assert "number of batches to wait" in str(excinfo.value)

    # Division
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, epochs=5)
    assert "should be a multiple of n_estimators" in str(excinfo.value)


def test_adversarial_training():
    model = torchensemble.AdversarialTrainingClassifier(
        estimator=MLP, n_estimators=2, cuda=False
    )

    # Epochs
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, epochs=-1)
    assert "number of training epochs" in str(excinfo.value)

    # Epsilon
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, epsilon=2)
    assert "step used to generate adversarial samples" in str(excinfo.value)

    # Log interval
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader, log_interval=-1)
    assert "number of batches to wait" in str(excinfo.value)



import torch
import pytest
import numpy as np
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader

import torchensemble
from torchensemble.utils import io
from torchensemble.utils.logging import set_logger


# All classifiers
all_clf = [
    torchensemble.FusionClassifier,
    torchensemble.VotingClassifier,
    torchensemble.BaggingClassifier,
    torchensemble.GradientBoostingClassifier,
    torchensemble.SnapshotEnsembleClassifier,
    torchensemble.AdversarialTrainingClassifier,
    torchensemble.FastGeometricClassifier,
    torchensemble.SoftGradientBoostingClassifier,
]


# All regressors
all_reg = [
    torchensemble.FusionRegressor,
    torchensemble.VotingRegressor,
    torchensemble.BaggingRegressor,
    torchensemble.GradientBoostingRegressor,
    torchensemble.SnapshotEnsembleRegressor,
    torchensemble.AdversarialTrainingRegressor,
    torchensemble.FastGeometricRegressor,
    torchensemble.SoftGradientBoostingRegressor,
]


np.random.seed(0)
torch.manual_seed(0)
device = torch.device("cpu")
logger = set_logger("pytest_all_models_multiple_input")


# Base estimator
class MLP_clf(nn.Module):
    def __init__(self):
        super(MLP_clf, self).__init__()
        self.linear1 = nn.Linear(2, 2)
        self.linear2 = nn.Linear(2, 2)

    def forward(self, X_1, X_2):
        X_1 = X_1.view(X_1.size()[0], -1)
        X_2 = X_2.view(X_2.size()[0], -1)
        output_1 = self.linear1(X_1)
        output_1 = self.linear2(output_1)
        output_2 = self.linear1(X_2)
        output_2 = self.linear2(output_2)
        return 0.5 * output_1 + 0.5 * output_2


class MLP_reg(nn.Module):
    def __init__(self):
        super(MLP_reg, self).__init__()
        self.linear1 = nn.Linear(2, 2)
        self.linear2 = nn.Linear(2, 1)

    def forward(self, X_1, X_2):
        X_1 = X_1.view(X_1.size()[0], -1)
        X_2 = X_2.view(X_2.size()[0], -1)
        output_1 = self.linear1(X_1)
        output_1 = self.linear2(output_1)
        output_2 = self.linear1(X_2)
        output_2 = self.linear2(output_2)
        return 0.5 * output_1 + 0.5 * output_2


# Training data
X_train = torch.Tensor(
    np.array(([0.1, 0.1], [0.2, 0.2], [0.3, 0.3], [0.4, 0.4]))
)

y_train_clf = torch.LongTensor(np.array(([0, 0, 1, 1])))
y_train_reg = torch.FloatTensor(np.array(([0.1, 0.2, 0.3, 0.4])))
y_train_reg = y_train_reg.view(-1, 1)


# Testing data
numpy_X_test = np.array(([0.5, 0.5], [0.6, 0.6]))
X_test = torch.Tensor(numpy_X_test)

y_test_clf = torch.LongTensor(np.array(([1, 0])))
y_test_reg = torch.FloatTensor(np.array(([0.5, 0.6])))
y_test_reg = y_test_reg.view(-1, 1)


@pytest.mark.parametrize("clf", all_clf)
def test_clf_class(clf):
    """
    This unit test checks the training and evaluating stage of all classifiers.
    """
    epochs = 1
    n_estimators = 2

    model = clf(estimator=MLP_clf, n_estimators=n_estimators, cuda=False)

    # Optimizer
    model.set_optimizer("Adam", lr=1e-3, weight_decay=5e-4)

    # Scheduler (Snapshot Ensemble Excluded)
    if not isinstance(model, torchensemble.SnapshotEnsembleClassifier):
        model.set_scheduler("MultiStepLR", milestones=[2, 4])

    # Prepare data with multiple inputs
    train = TensorDataset(X_train, X_train, y_train_clf)
    train_loader = DataLoader(train, batch_size=2, shuffle=False)
    test = TensorDataset(X_test, X_test, y_test_clf)
    test_loader = DataLoader(test, batch_size=2, shuffle=False)

    # Snapshot Ensemble needs more epochs
    if isinstance(model, torchensemble.SnapshotEnsembleClassifier):
        epochs = 6

    # Train
    model.fit(train_loader, epochs=epochs, test_loader=test_loader)

    # Evaluate
    model.evaluate(test_loader)

    # Predict
    for _, elem in enumerate(test_loader):
        data, target = io.split_data_target(elem, device)
        model.predict(*data)
        break

    # Reload
    new_model = clf(estimator=MLP_clf, n_estimators=n_estimators, cuda=False)
    io.load(new_model)

    new_model.evaluate(test_loader)

    for _, elem in enumerate(test_loader):
        data, target = io.split_data_target(elem, device)
        new_model.predict(*data)
        break


@pytest.mark.parametrize("clf", all_clf)
def test_clf_object(clf):
    """
    This unit test checks the training and evaluating stage of all classifiers.
    """
    epochs = 1
    n_estimators = 2

    model = clf(estimator=MLP_clf(), n_estimators=n_estimators, cuda=False)

    # Optimizer
    model.set_optimizer("Adam", lr=1e-3, weight_decay=5e-4)

    # Scheduler (Snapshot Ensemble Excluded)
    if not isinstance(model, torchensemble.SnapshotEnsembleClassifier):
        model.set_scheduler("MultiStepLR", milestones=[2, 4])

    # Prepare data with multiple inputs
    train = TensorDataset(X_train, X_train, y_train_clf)
    train_loader = DataLoader(train, batch_size=2, shuffle=False)
    test = TensorDataset(X_test, X_test, y_test_clf)
    test_loader = DataLoader(test, batch_size=2, shuffle=False)

    # Snapshot Ensemble needs more epochs
    if isinstance(model, torchensemble.SnapshotEnsembleClassifier):
        epochs = 6

    # Train
    model.fit(train_loader, epochs=epochs, test_loader=test_loader)

    # Evaluate
    model.evaluate(test_loader)

    # Predict
    for _, elem in enumerate(test_loader):
        data, target = io.split_data_target(elem, device)
        model.predict(*data)
        break

    # Reload
    new_model = clf(estimator=MLP_clf(), n_estimators=n_estimators, cuda=False)
    io.load(new_model)

    new_model.evaluate(test_loader)

    for _, elem in enumerate(test_loader):
        data, target = io.split_data_target(elem, device)
        new_model.predict(*data)
        break


@pytest.mark.parametrize("reg", all_reg)
def test_reg_class(reg):
    """
    This unit test checks the training and evaluating stage of all regressors.
    """
    epochs = 1
    n_estimators = 2

    model = reg(estimator=MLP_reg, n_estimators=n_estimators, cuda=False)

    # Optimizer
    model.set_optimizer("Adam", lr=1e-3, weight_decay=5e-4)

    # Scheduler (Snapshot Ensemble Excluded)
    if not isinstance(model, torchensemble.SnapshotEnsembleRegressor):
        model.set_scheduler("MultiStepLR", milestones=[2, 4])

    # Prepare data with multiple inputs
    train = TensorDataset(X_train, X_train, y_train_reg)
    train_loader = DataLoader(train, batch_size=2, shuffle=False)
    test = TensorDataset(X_test, X_test, y_test_reg)
    test_loader = DataLoader(test, batch_size=2, shuffle=False)

    # Snapshot Ensemble needs more epochs
    if isinstance(model, torchensemble.SnapshotEnsembleRegressor):
        epochs = 6

    # Train
    model.fit(train_loader, epochs=epochs, test_loader=test_loader)

    # Evaluate
    model.evaluate(test_loader)

    # Predict
    for _, elem in enumerate(test_loader):
        data, target = io.split_data_target(elem, device)
        model.predict(*data)
        break

    # Reload
    new_model = reg(estimator=MLP_reg, n_estimators=n_estimators, cuda=False)
    io.load(new_model)

    new_model.evaluate(test_loader)

    for _, elem in enumerate(test_loader):
        data, target = io.split_data_target(elem, device)
        new_model.predict(*data)
        break


@pytest.mark.parametrize("reg", all_reg)
def test_reg_object(reg):
    """
    This unit test checks the training and evaluating stage of all regressors.
    """
    epochs = 1
    n_estimators = 2

    model = reg(estimator=MLP_reg(), n_estimators=n_estimators, cuda=False)

    # Optimizer
    model.set_optimizer("Adam", lr=1e-3, weight_decay=5e-4)

    # Scheduler (Snapshot Ensemble Excluded)
    if not isinstance(model, torchensemble.SnapshotEnsembleRegressor):
        model.set_scheduler("MultiStepLR", milestones=[2, 4])

    # Prepare data with multiple inputs
    train = TensorDataset(X_train, X_train, y_train_reg)
    train_loader = DataLoader(train, batch_size=2, shuffle=False)
    test = TensorDataset(X_test, X_test, y_test_reg)
    test_loader = DataLoader(test, batch_size=2, shuffle=False)

    # Snapshot Ensemble needs more epochs
    if isinstance(model, torchensemble.SnapshotEnsembleRegressor):
        epochs = 6

    # Train
    model.fit(train_loader, epochs=epochs, test_loader=test_loader)

    # Evaluate
    model.evaluate(test_loader)

    # Predict
    for _, elem in enumerate(test_loader):
        data, target = io.split_data_target(elem, device)
        model.predict(*data)
        break

    # Reload
    new_model = reg(estimator=MLP_reg(), n_estimators=n_estimators, cuda=False)
    io.load(new_model)

    new_model.evaluate(test_loader)

    for _, elem in enumerate(test_loader):
        data, target = io.split_data_target(elem, device)
        new_model.predict(*data)
        break


def test_split_data_target_invalid_data_type():
    with pytest.raises(ValueError) as excinfo:
        io.split_data_target(0.0, device, logger)
    assert "Invalid dataloader" in str(excinfo.value)


def test_split_data_target_invalid_list_length():
    with pytest.raises(ValueError) as excinfo:
        io.split_data_target([0.0], device, logger)
    assert "should at least contain two tensors" in str(excinfo.value)



"""

Reeds Shepp path planner sample code

author Atsushi Sakai(@Atsushi_twi)

"""
import math

import matplotlib.pyplot as plt
import numpy as np

show_animation = True


class Path:
    """
    Path data container
    """

    def __init__(self):
        # course segment length  (negative value is backward segment)
        self.lengths = []
        # course segment type char ("S": straight, "L": left, "R": right)
        self.ctypes = []
        self.L = 0.0  # Total lengths of the path
        self.x = []  # x positions
        self.y = []  # y positions
        self.yaw = []  # orientations [rad]
        self.directions = []  # directions (1:forward, -1:backward)


def plot_arrow(x, y, yaw, length=1.0, width=0.5, fc="r", ec="k"):
    if isinstance(x, list):
        for (ix, iy, iyaw) in zip(x, y, yaw):
            plot_arrow(ix, iy, iyaw)
    else:
        plt.arrow(x, y, length * math.cos(yaw), length * math.sin(yaw), fc=fc,
                  ec=ec, head_width=width, head_length=width)
        plt.plot(x, y)


def mod2pi(x):
    # Be consistent with fmod in cplusplus here.
    v = np.mod(x, np.copysign(2.0 * math.pi, x))
    if v < -math.pi:
        v += 2.0 * math.pi
    else:
        if v > math.pi:
            v -= 2.0 * math.pi
    return v


def straight_left_straight(x, y, phi):
    phi = mod2pi(phi)
    if y > 0.0 and 0.0 < phi < math.pi * 0.99:
        xd = - y / math.tan(phi) + x
        t = xd - math.tan(phi / 2.0)
        u = phi
        v = math.sqrt((x - xd) ** 2 + y ** 2) - math.tan(phi / 2.0)
        return True, t, u, v
    elif y < 0.0 < phi < math.pi * 0.99:
        xd = - y / math.tan(phi) + x
        t = xd - math.tan(phi / 2.0)
        u = phi
        v = -math.sqrt((x - xd) ** 2 + y ** 2) - math.tan(phi / 2.0)
        return True, t, u, v

    return False, 0.0, 0.0, 0.0


def set_path(paths, lengths, ctypes, step_size):
    path = Path()
    path.ctypes = ctypes
    path.lengths = lengths
    path.L = sum(np.abs(lengths))

    # check same path exist
    for i_path in paths:
        type_is_same = (i_path.ctypes == path.ctypes)
        length_is_close = (sum(np.abs(i_path.lengths)) - path.L) <= step_size
        if type_is_same and length_is_close:
            return paths  # same path found, so do not insert path

    # check path is long enough
    if path.L <= step_size:
        return paths  # too short, so do not insert path

    paths.append(path)
    return paths


def straight_curve_straight(x, y, phi, paths, step_size):
    flag, t, u, v = straight_left_straight(x, y, phi)
    if flag:
        paths = set_path(paths, [t, u, v], ["S", "L", "S"], step_size)

    flag, t, u, v = straight_left_straight(x, -y, -phi)
    if flag:
        paths = set_path(paths, [t, u, v], ["S", "R", "S"], step_size)

    return paths


def polar(x, y):
    r = math.sqrt(x ** 2 + y ** 2)
    theta = math.atan2(y, x)
    return r, theta


def left_straight_left(x, y, phi):
    u, t = polar(x - math.sin(phi), y - 1.0 + math.cos(phi))
    if t >= 0.0:
        v = mod2pi(phi - t)
        if v >= 0.0:
            return True, t, u, v

    return False, 0.0, 0.0, 0.0


def left_right_left(x, y, phi):
    u1, t1 = polar(x - math.sin(phi), y - 1.0 + math.cos(phi))

    if u1 <= 4.0:
        u = -2.0 * math.asin(0.25 * u1)
        t = mod2pi(t1 + 0.5 * u + math.pi)
        v = mod2pi(phi - t + u)

        if t >= 0.0 >= u:
            return True, t, u, v

    return False, 0.0, 0.0, 0.0


def curve_curve_curve(x, y, phi, paths, step_size):
    flag, t, u, v = left_right_left(x, y, phi)
    if flag:
        paths = set_path(paths, [t, u, v], ["L", "R", "L"], step_size)

    flag, t, u, v = left_right_left(-x, y, -phi)
    if flag:
        paths = set_path(paths, [-t, -u, -v], ["L", "R", "L"], step_size)

    flag, t, u, v = left_right_left(x, -y, -phi)
    if flag:
        paths = set_path(paths, [t, u, v], ["R", "L", "R"], step_size)

    flag, t, u, v = left_right_left(-x, -y, phi)
    if flag:
        paths = set_path(paths, [-t, -u, -v], ["R", "L", "R"], step_size)

    # backwards
    xb = x * math.cos(phi) + y * math.sin(phi)
    yb = x * math.sin(phi) - y * math.cos(phi)

    flag, t, u, v = left_right_left(xb, yb, phi)
    if flag:
        paths = set_path(paths, [v, u, t], ["L", "R", "L"], step_size)

    flag, t, u, v = left_right_left(-xb, yb, -phi)
    if flag:
        paths = set_path(paths, [-v, -u, -t], ["L", "R", "L"], step_size)

    flag, t, u, v = left_right_left(xb, -yb, -phi)
    if flag:
        paths = set_path(paths, [v, u, t], ["R", "L", "R"], step_size)

    flag, t, u, v = left_right_left(-xb, -yb, phi)
    if flag:
        paths = set_path(paths, [-v, -u, -t], ["R", "L", "R"], step_size)

    return paths


def curve_straight_curve(x, y, phi, paths, step_size):
    flag, t, u, v = left_straight_left(x, y, phi)
    if flag:
        paths = set_path(paths, [t, u, v], ["L", "S", "L"], step_size)

    flag, t, u, v = left_straight_left(-x, y, -phi)
    if flag:
        paths = set_path(paths, [-t, -u, -v], ["L", "S", "L"], step_size)

    flag, t, u, v = left_straight_left(x, -y, -phi)
    if flag:
        paths = set_path(paths, [t, u, v], ["R", "S", "R"], step_size)

    flag, t, u, v = left_straight_left(-x, -y, phi)
    if flag:
        paths = set_path(paths, [-t, -u, -v], ["R", "S", "R"], step_size)

    flag, t, u, v = left_straight_right(x, y, phi)
    if flag:
        paths = set_path(paths, [t, u, v], ["L", "S", "R"], step_size)

    flag, t, u, v = left_straight_right(-x, y, -phi)
    if flag:
        paths = set_path(paths, [-t, -u, -v], ["L", "S", "R"], step_size)

    flag, t, u, v = left_straight_right(x, -y, -phi)
    if flag:
        paths = set_path(paths, [t, u, v], ["R", "S", "L"], step_size)

    flag, t, u, v = left_straight_right(-x, -y, phi)
    if flag:
        paths = set_path(paths, [-t, -u, -v], ["R", "S", "L"], step_size)

    return paths


def left_straight_right(x, y, phi):
    u1, t1 = polar(x + math.sin(phi), y - 1.0 - math.cos(phi))
    u1 = u1 ** 2
    if u1 >= 4.0:
        u = math.sqrt(u1 - 4.0)
        theta = math.atan2(2.0, u)
        t = mod2pi(t1 + theta)
        v = mod2pi(t - phi)

        if t >= 0.0 and v >= 0.0:
            return True, t, u, v

    return False, 0.0, 0.0, 0.0


def generate_path(q0, q1, max_curvature, step_size):
    dx = q1[0] - q0[0]
    dy = q1[1] - q0[1]
    dth = q1[2] - q0[2]
    c = math.cos(q0[2])
    s = math.sin(q0[2])
    x = (c * dx + s * dy) * max_curvature
    y = (-s * dx + c * dy) * max_curvature

    paths = []
    paths = straight_curve_straight(x, y, dth, paths, step_size)
    paths = curve_straight_curve(x, y, dth, paths, step_size)
    paths = curve_curve_curve(x, y, dth, paths, step_size)

    return paths


def calc_interpolate_dists_list(lengths, step_size):
    interpolate_dists_list = []
    for length in lengths:
        d_dist = step_size if length >= 0.0 else -step_size
        interp_dists = np.arange(0.0, length, d_dist)
        interp_dists = np.append(interp_dists, length)
        interpolate_dists_list.append(interp_dists)

    return interpolate_dists_list


def generate_local_course(lengths, modes, max_curvature, step_size):
    interpolate_dists_list = calc_interpolate_dists_list(lengths, step_size)

    origin_x, origin_y, origin_yaw = 0.0, 0.0, 0.0

    xs, ys, yaws, directions = [], [], [], []
    for (interp_dists, mode, length) in zip(interpolate_dists_list, modes,
                                            lengths):

        for dist in interp_dists:
            x, y, yaw, direction = interpolate(dist, length, mode,
                                               max_curvature, origin_x,
                                               origin_y, origin_yaw)
            xs.append(x)
            ys.append(y)
            yaws.append(yaw)
            directions.append(direction)
        origin_x = xs[-1]
        origin_y = ys[-1]
        origin_yaw = yaws[-1]

    return xs, ys, yaws, directions


def interpolate(dist, length, mode, max_curvature, origin_x, origin_y,
                origin_yaw):
    if mode == "S":
        x = origin_x + dist / max_curvature * math.cos(origin_yaw)
        y = origin_y + dist / max_curvature * math.sin(origin_yaw)
        yaw = origin_yaw
    else:  # curve
        ldx = math.sin(dist) / max_curvature
        ldy = 0.0
        yaw = None
        if mode == "L":  # left turn
            ldy = (1.0 - math.cos(dist)) / max_curvature
            yaw = origin_yaw + dist
        elif mode == "R":  # right turn
            ldy = (1.0 - math.cos(dist)) / -max_curvature
            yaw = origin_yaw - dist
        gdx = math.cos(-origin_yaw) * ldx + math.sin(-origin_yaw) * ldy
        gdy = -math.sin(-origin_yaw) * ldx + math.cos(-origin_yaw) * ldy
        x = origin_x + gdx
        y = origin_y + gdy

    return x, y, yaw, 1 if length > 0.0 else -1


def pi_2_pi(angle):
    return (angle + math.pi) % (2 * math.pi) - math.pi


def calc_paths(sx, sy, syaw, gx, gy, gyaw, maxc, step_size):
    q0 = [sx, sy, syaw]
    q1 = [gx, gy, gyaw]

    paths = generate_path(q0, q1, maxc, step_size)
    
    # bug fix
    for i, path in enumerate(paths):
        for le in path.lengths:
            if le > 1000 or le < -1000:
                del paths[i]
    
    for path in paths:
        xs, ys, yaws, directions = generate_local_course(path.lengths,
                                                         path.ctypes, maxc,
                                                         step_size * maxc)

        # convert global coordinate
        path.x = [math.cos(-q0[2]) * ix + math.sin(-q0[2]) * iy + q0[0] for
                  (ix, iy) in zip(xs, ys)]
        path.y = [-math.sin(-q0[2]) * ix + math.cos(-q0[2]) * iy + q0[1] for
                  (ix, iy) in zip(xs, ys)]
        path.yaw = [pi_2_pi(yaw + q0[2]) for yaw in yaws]
        path.directions = directions
        path.lengths = [length / maxc for length in path.lengths]
        path.L = path.L / maxc

    return paths


def reeds_shepp_path_planning(sx, sy, syaw, gx, gy, gyaw, maxc, step_size=0.2):
    paths = calc_paths(sx, sy, syaw, gx, gy, gyaw, maxc, step_size)
    if not paths:
        return None, None, None, None, None  # could not generate any path

    # search minimum cost path
    best_path_index = paths.index(min(paths, key=lambda p: abs(p.L)))
    b_path = paths[best_path_index]

    return b_path.x, b_path.y, b_path.yaw, b_path.ctypes, b_path.lengths


def main():
    print("Reeds Shepp path planner sample start!!")

    start_x = -1.0  # [m]
    start_y = -4.0  # [m]
    start_yaw = np.deg2rad(-20.0)  # [rad]

    end_x = 5.0  # [m]
    end_y = 5.0  # [m]
    end_yaw = np.deg2rad(25.0)  # [rad]

    curvature = 0.1
    step_size = 0.05

    xs, ys, yaws, modes, lengths = reeds_shepp_path_planning(start_x, start_y,
                                                             start_yaw, end_x,
                                                             end_y, end_yaw,
                                                             curvature,
                                                             step_size)

    if show_animation:  # pragma: no cover
        plt.cla()
        plt.plot(xs, ys, label="final course " + str(modes))
        print(f"{lengths=}")

        # plotting
        plot_arrow(start_x, start_y, start_yaw)
        plot_arrow(end_x, end_y, end_yaw)

        plt.legend()
        plt.grid(True)
        plt.axis("equal")
        plt.show()

    if not xs:
        assert False, "No path"


if __name__ == '__main__':
    main()



from matplotlib import pyplot as plt

from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.configs import *


# Set stats to true to track bullet stats
cfg = MasterConfig()
cfg.env.general.gui = True
cfg.maps.general.map_name = 'narrow'
cfg.maps.narrow.narrow1_pos = [3.5, 3.5]
cfg.maps.narrow.narrow2_pos = [7, 5.5]
cfg.env.car.pose = [1.5, 6.0, 1.2]
cfg.env.target.pose = [8.5, 4.0, -0.6]

# Reset master to track stats
env = CarNavigationBulletEnv(cfg)
env.reset()
img = env._get_image_current_top_view()
plt.imshow(img, origin="lower", extent=env.img_extent)
plt.savefig('narrow_top_view.png', dpi=600)


import mrl
import numpy as np
import gym
from mrl.replays.core.replay_buffer import RingBuffer
from mrl.utils.misc import AttrDict
import pickle
import os
from sklearn import mixture
from scipy.stats import rankdata

class EntropyPrioritizedOnlineHERBuffer(mrl.Module):

  def __init__(
      self,
      module_name='prioritized_replay',
      rank_method='dense',
      temperature=1.0
  ):
    """
    Buffer that stores entropy of trajectories for prioritized replay
    """

    super().__init__(module_name, required_agent_modules=['env','replay_buffer'], locals=locals())

    self.goal_space = None
    self.buffer = None
    self.rank_method = rank_method
    self.temperature = temperature
    self.traj_len = None
    self.has_fit_density = False

  def _setup(self):
    self.ag_buffer = self.replay_buffer.buffer.BUFF.buffer_ag

    env = self.env
    assert type(env.observation_space) == gym.spaces.Dict
    self.goal_space = env.observation_space.spaces["desired_goal"]

    # Note: for now we apply entropy estimation on the achieved goal (ag) space
    # Define the buffers to store for prioritization
    items = [("entropy", (1,)), ("priority", (1,))]
    self.buffer = AttrDict()
    for name, shape in items:
      self.buffer['buffer_' + name] = RingBuffer(self.ag_buffer.maxlen, shape=shape)

    self._subbuffers = [[] for _ in range(self.env.num_envs)]
    self.n_envs = self.env.num_envs

    # Define the placeholder for mixture model to estimate trajectory
    self.clf = 0

  def fit_density_model(self):
    if not self.has_fit_density:
      self.has_fit_density = True
    ag = self.ag_buffer.data[0:self.size].copy()
    X_train = ag.reshape(-1, self.traj_len * ag.shape[-1]) # [num_episodes, episode_len * goal_dim]

    self.clf = mixture.BayesianGaussianMixture(weight_concentration_prior_type="dirichlet_distribution", n_components=3)
    self.clf.fit(X_train)
    pred = -self.clf.score_samples(X_train)

    self.pred_min = pred.min()
    pred = pred - self.pred_min
    pred = np.clip(pred, 0, None)
    self.pred_sum = pred.sum()
    pred = pred / self.pred_sum
    self.pred_avg = (1 / pred.shape[0])
    pred = np.repeat(pred, self.traj_len, axis=0)

    self.buffer.buffer_entropy.data[:self.size] = pred.reshape(-1,1).copy()

  def _process_experience(self, exp):
    # Compute the entropy 
    # TODO: Include previous achieved goal too? or use that instead of ag?
    achieved = exp.next_state['achieved_goal']
    for i in range(self.n_envs):
      self._subbuffers[i].append([achieved[i]])
    
    for i in range(self.n_envs):
      if exp.trajectory_over[i]:
        # TODO: Compute the entropy of the trajectory
        traj_len = len(self._subbuffers[i])
        if self.traj_len is None:
          self.traj_len = traj_len
        else:
          # Current implementation assumes the same length for all trajectories
          assert(traj_len == self.traj_len)

        if not isinstance(self.clf, int):
          ag = [np.stack(a) for a in zip(*self._subbuffers[i])][0] # [episode_len, goal_dim]
          X = ag.reshape(-1, ag.shape[0]*ag.shape[1])
          pred = -self.clf.score_samples(X)

          pred = pred - self.pred_min
          pred = np.clip(pred, 0, None)
          pred = pred / self.pred_sum # Shape (1,)

          entropy = np.ones((traj_len,1)) * pred
        else:
          # Not enough data to train mixture density yet, set entropy to be zero
          entropy = np.zeros((traj_len, 1))
        
        priority = np.zeros((traj_len,1))
        trajectory = [entropy, priority]
        
        # TODO: Update the trajectory with entropy
        self.add_trajectory(*trajectory)

        self._subbuffers[i] = []

        # TODO: Update the rank here before adding it to the trajectory?
        self.update_priority()

  def add_trajectory(self, *items):
    """
    Append a trajectory of transitions to the buffer.

    :param items: a list of batched transition values to append to the replay buffer,
        in the item order that we initialized the ReplayBuffer with.
    """
    for buffer, batched_values in zip(self.buffer.values(), items):
      buffer.append_batch(batched_values)

  def update_priority(self):
    """
    After adding a trajectory to the replay buffer, update the ranking of transitions
    """
    # Note: 'dense' assigns the next highest element with the rank immediately 
    # after those assigned to the tied elements.
    entropy_transition_total = self.buffer.buffer_entropy.data[:self.size]
    entropy_rank = rankdata(entropy_transition_total, method=self.rank_method)
    entropy_rank = (entropy_rank - 1).reshape(-1, 1)
    self.buffer.buffer_priority.data[:self.size] = entropy_rank

  def __call__(self, batch_size):
    """
    Samples batch_size number of indices from main replay_buffer.

    Args:
      batch_size (int): size of the batch to sample
    
    Returns:
      batch_idxs: a 1-D numpy array of length batch_size containing indices
                  sampled in prioritized manner
    """
    if self.rank_method == 'none':
      entropy_trajectory = self.buffer.buffer_entropy.data[:self.size]
    else:
      entropy_trajectory = self.buffer.buffer_priority.data[:self.size]
    
    # Factorize out sampling into sampling trajectory according to priority/entropy
    # then sample time uniformly independently
    entropy_trajectory = entropy_trajectory.reshape(-1, self.traj_len)[:,0]
    p_trajectory = np.power(entropy_trajectory, 1/(self.temperature+1e-2))
    
    # If the density model hasn't been fitted yet, we have p_trajectory all 0's
    # And hence treat them as uniform:
    if not self.has_fit_density:
      p_trajectory = np.ones(p_trajectory.shape) / len(p_trajectory)
    else:
      assert(p_trajectory.sum() != 0.0)
      p_trajectory = p_trajectory / p_trajectory.sum()
    
    num_trajectories = p_trajectory.shape[0]
    batch_tidx = np.random.choice(num_trajectories, size=batch_size, p=p_trajectory)
    batch_idxs = self.traj_len * batch_tidx + np.random.choice(self.traj_len, size=batch_size)

    return batch_idxs

  @property
  def size(self):
    return len(self.ag_buffer)

  def save(self, save_folder):
    if self.config.save_replay_buf:
      state = self.buffer._get_state()
      with open(os.path.join(save_folder, "{}.pickle".format(self.module_name)), 'wb') as f:
        pickle.dump(state, f)

  def load(self, save_folder):
    load_path = os.path.join(save_folder, "{}.pickle".format(self.module_name))
    if os.path.exists(load_path):
      with open(load_path, 'rb') as f:
        state = pickle.load(f)
      self.buffer._set_state(state)
    else:
      self.logger.log_color('###############################################################', '', color='red')
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='cyan')
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='red')
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='yellow')
      self.logger.log_color('###############################################################', '', color='red')


"""
Success Prediction Module
"""

import mrl
import torch, torch.nn as nn, torch.nn.functional as F
import numpy as np
import os
from mrl.replays.online_her_buffer import OnlineHERBuffer

class GoalSuccessPredictor(mrl.Module):
  """Predicts success using a learned discriminator"""

  def __init__(self, batch_size = 50, history_length = 200, optimize_every=250, log_every=5000):
    super().__init__(
      'success_predictor',
      required_agent_modules=[
        'env', 'replay_buffer', 'goal_discriminator'
      ],
      locals=locals())
    self.log_every = log_every
    self.batch_size = batch_size
    self.history_length = history_length
    self.optimize_every = optimize_every
    self.opt_steps = 0

  def _setup(self):
    super()._setup()
    assert isinstance(self.replay_buffer, OnlineHERBuffer)
    assert self.env.goal_env
    self.n_envs = self.env.num_envs
    self.optimizer = torch.optim.Adam(self.goal_discriminator.model.parameters())

  def _optimize(self):
    self.opt_steps += 1

    if len(self.replay_buffer.buffer.trajectories) > self.batch_size and self.opt_steps % self.optimize_every == 0:
      trajs = self.replay_buffer.buffer.sample_trajectories(self.batch_size, group_by_buffer=True, from_m_most_recent=self.history_length)
      successes = np.array([np.any(np.isclose(traj, 0.), axis=0) for traj in trajs[2]])

      start_states = np.array([t[0] for t in trajs[0]])
      behav_goals =  np.array([t[0] for t in trajs[7]])
      states = np.concatenate((start_states, behav_goals), -1)

      targets = self.torch(successes)
      inputs = self.torch(states)

      # outputs here have not been passed through sigmoid
      outputs = self.goal_discriminator(inputs)
      loss = F.binary_cross_entropy_with_logits(outputs, targets)

      if hasattr(self, 'logger'):
        self.logger.add_histogram('predictions', torch.sigmoid(outputs), self.log_every)
        self.logger.add_histogram('targets', targets, self.log_every)

      # optimize
      self.optimizer.zero_grad()
      loss.backward()
      self.optimizer.step()

  def __call__(self, *states_and_maybe_goals):
    """Input / output are numpy arrays"""
    states = np.concatenate(states_and_maybe_goals, -1)
    return self.numpy(torch.sigmoid(self.goal_discriminator(self.torch(states))))

  def save(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    torch.save({
      'opt_state_dict': self.optimizer.state_dict()
    }, path)

  def load(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    checkpoint = torch.load(path)
    self.optimizer.load_state_dict(checkpoint['opt_state_dict'])
